// Copyright Contributors to the OpenVDB Project
// SPDX-License-Identifier: MPL-2.0

/*!
    @file AlignedHostBuffer.h

    @brief AlignedHostBuffer contains modified code from HostBuffer.h.
           Implements an aligned buffer that contains a shared or private bump
           pool to either externally or internally managed host memory.

    @details This AlignedHostBuffer can be used in multiple ways, most of which are
             demonstrated in the examples below. Memory in the pool can
             be managed or unmanged (e.g. internal or external) and can
             be shared between multiple buffers or belong to a single buffer.

   Example that uses AlignedHostBuffer::create inside io::readGrids to create a
   full self-managed buffer, i.e. not shared and without padding, per grid in the file.
   @code
        auto handles = nanovdb::io::readGrids("file.nvdb");
   @endcode

   Example that uses AlignedHostBuffer::createFull. Assuming you have a raw pointer
   to a NanoVDB grid of unknown type, this examples shows how to create its
   GridHandle which can be used to enquire about the grid type and meta data.
   @code
        void    *data;// pointer to a NanoVDB grid of unknown type
        uint64_t size;// byte size of NanoVDB grid of unknown type
        auto buffer = nanovdb::AlignedHostBuffer::createFull(size, data);
        nanovdb::GridHandle<> gridHandle(std::move(buffer));
   @endcode

   Example that uses AlignedHostBuffer::createPool for internally managed host memory.
   Suppose you want to read multiple grids in multiple files, but reuse the same
   fixed sized memory buffer to both avoid memory fragmentation as well as
   exceeding the fixed memory ceiling!
   @code
        auto pool = nanovdb::AlignedHostBuffer::createPool(1 << 30);// 1 GB memory pool
        std::vector<std::string>> frames;// vector of grid names
        for (int i=0; i<frames.size(); ++i) {
            auto handles = nanovdb::io::readGrids(frames[i], 0, pool);// throws if grids in file exceed 1 GB
            ...
            pool.reset();// clears all handles and resets the memory pool for reuse
        }
   @endcode

   Example that uses AlignedHostBuffer::createPool for externally managed host memory.
   Note that in this example @c handles are allowed to outlive @c pool since
   they internally store a shared pointer to the memory pool. However @c data
   MUST outlive @c handles since the pool does not own its memory in this example.
   @code
        const size_t poolSize = 1 << 30;// 1 GB
        uint8_t *data = static_cast<uint8_t*>(std::malloc(size));// 1 GB buffer
        auto pool = nanovdb::AlignedHostBuffer::createPool(poolSize, data);
        auto handles1 = nanovdb::io::readGrids("file1.nvdb", 0, pool);
        auto handles2 = nanovdb::io::readGrids("file2.nvdb", 0, pool);
        ....
        std::free(data);
   @endcode

   Example that uses AlignedHostBuffer::createPool for externally managed host memory.
   Note that in this example @c handles are allowed to outlive @c pool since
   they internally store a shared pointer to the memory pool. However @c array
   MUST outlive @c handles since the pool does not own its memory in this example.
   @code
        const size_t poolSize = 1 << 30;// 1 GB
        std::unique_ptr<uint8_t[]> array(new uint8_t[size]);// scoped buffer of 1 GB
        auto pool = nanovdb::AlignedHostBuffer::createPool(poolSize, array.get());
        auto handles = nanovdb::io::readGrids("file.nvdb", 0, pool);
   @endcode
*/

#ifndef NANOVTT_ALIGNEDHOSTBUFFER_H_HAS_BEEN_INCLUDED
#define NANOVTT_ALIGNEDHOSTBUFFER_H_HAS_BEEN_INCLUDED

#include <stdint.h> //    for types like int32_t etc
#include <cstdio> //      for fprintf
#include <cstdlib> //     for std::malloc/std::reallow/std::free
#include <memory>//       for std::make_shared
#include <mutex>//        for std::mutex
#include <unordered_set>//for std::unordered_set
#include <cassert>//      for assert
#include <sstream>//      for std::stringstream
#include <cstring>//      for memcpy

#if defined(__clang__)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wunused-parameter"
#pragma clang diagnostic ignored "-Wold-style-cast"
#endif

#include <Bifrost/GeometryVolume/internal/nanovdb/util/HostBuffer.h>// for BufferTraits

#include <Bifrost/GeometryVolume/nanovtt/NanoVTT.h>

#define checkPtr(ptr, msg) \
    { \
        ptrAssert((ptr), (msg), __FILE__, __LINE__); \
    }

namespace nanovtt {


// ----------------------------> AlignedHostBuffer <--------------------------------------

/// @brief This is a buffer that contains a shared or private pool
///        to either externally or internally managed host memory.
///
/// @note  Terminology:
///        Pool:   0 = buffer.size() < buffer.poolSize()
///        Buffer: 0 < buffer.size() < buffer.poolSize()
///        Full:   0 < buffer.size() = buffer.poolSize()
///        Empty:  0 = buffer.size() = buffer.poolSize()
class AlignedHostBuffer
{
    struct Pool;// forward declaration of private pool struct
    std::shared_ptr<Pool> mPool;
    uint64_t              mSize; // total number of bytes for the NanoVDB grid.
    uint8_t*              mData; // raw buffer for the NanoVDB grid.

#if defined(DEBUG) || defined(_DEBUG)
    static inline void ptrAssert(void* ptr, const char* msg, const char* file, int line, bool abort = true)
    {
        if (ptr == nullptr) {
            fprintf(stderr, "NULL pointer error: %s %s %d\n", msg, file, line);
            if (abort)
                exit(1);
        }
    }
#else
    static inline void ptrAssert(void*, const char*, const char*, int, bool = true)
    {
    }
#endif

public:
    /// @brief Return a full buffer or an empty buffer
    AlignedHostBuffer(uint64_t bufferSize = 0);

     /// @brief Move copy-constructor
    AlignedHostBuffer(AlignedHostBuffer&& other);

    /// @brief Custom descructor
    ~AlignedHostBuffer() { this->clear(); }

    /// @brief Move copy assignment operation
    AlignedHostBuffer& operator=(AlignedHostBuffer&& other);

    /// @brief Disallow copy-construction
    AlignedHostBuffer(const AlignedHostBuffer&) = delete;

    /// @brief Disallow copy assignment operation
    AlignedHostBuffer& operator=(const AlignedHostBuffer&) = delete;

    /// @brief Return a pool buffer which satisfies: buffer.size == 0,
    ///        buffer.poolSize() == poolSize, and buffer.data() == nullptr.
    ///        If data==nullptr, memory for the pool will be allocated.
    ///
    /// @throw If poolSize is zero.
    static AlignedHostBuffer createPool(uint64_t poolSize, void *data = nullptr);

    /// @brief Return a full buffer which satisfies: buffer.size == bufferSize,
    ///        buffer.poolSize() == bufferSize, and buffer.data() == data.
    ///        If data==nullptr, memory for the pool will be allocated.
    ///
    /// @throw If bufferSize is zero.
    static AlignedHostBuffer createFull(uint64_t bufferSize, void *data = nullptr);

    /// @brief Return a buffer with @c bufferSize bytes managed by
    ///        the specified memory @c pool. If none is provided, i.e.
    ///        @c pool == nullptr or @c pool->poolSize() == 0, one is
    ///        created with size @c bufferSize, i.e. a full buffer is returned.
    ///
    /// @throw If the specified @c pool has insufficient memory for
    ///        the requested buffer size.
    static AlignedHostBuffer create(uint64_t bufferSize, const AlignedHostBuffer* pool = nullptr);

    /// @brief Initialize as a full buffer with the specified size. If data is NULL
    ///        the memory is internally allocated.
    void init(uint64_t bufferSize, void *data = nullptr);

    //@{
    /// @brief Retuns a pointer to the raw memory buffer managed by this allocator.
    ///
    /// @warning Note that the pointer can be NULL if the allocator was not initialized!
    const uint8_t* data() const { return mData; }
    uint8_t* data() { return mData; }
    uint8_t* deviceData() const { return nullptr; }
    //@}

    //@{
    /// NOP
    void deviceUpload(void* stream = 0, bool sync = true) const {}
    void deviceDownload(void* stream = 0, bool sync = true) const {}
    //@}

    //@{
    /// @brief Returns the size in bytes associated with this buffer.
    uint64_t bufferSize() const { return mSize; }
    uint64_t size() const { return this->bufferSize(); }
    //@}

    /// @brief Returns the size in bytes of the memory pool shared with this instance.
    uint64_t poolSize() const;

    /// @brief Return true if memory is managed (using std::malloc and std:free) by the
    ///        shared pool in this buffer. Else memory is assumed to be managed externally.
    bool isManaged() const;

    //@{
    /// @brief Returns true if this buffer has no memory associated with it
    bool isEmpty() const { return !mPool || mSize == 0 || mData == nullptr; }
    bool empty() const { return this->isEmpty(); }
    //@}

    /// @brief Return true if this is a pool, i.e. an empty buffer with a nonempty
    ///        internal pool, i.e. this->size() == 0 and this->poolSize() != 0
    bool isPool() const { return mSize == 0 && this->poolSize() > 0; }

    /// @brief Return true if the pool exists, is nonempty but has no more available memory
    bool isFull() const;

    /// @brief Clear this buffer so it is empty.
    void clear();

    /// @brief Clears all existing buffers that are registered against the memory pool
    ///        and resets the pool so it can be reused to create new buffers.
    ///
    /// @throw If this instance is not empty or contains no pool.
    ///
    /// @warning This method is not thread-safe!
    void reset();

    /// @brief Total number of bytes from the pool currently in use by buffers
    uint64_t poolUsage() const;

}; // AlignedHostBuffer class

// --------------------------> Implementation of AlignedHostBuffer::Pool <------------------------------------

// This is private struct of AlignedHostBuffer so you can safely ignore the API
struct AlignedHostBuffer::Pool
{
    using HashTableT = std::unordered_set<AlignedHostBuffer*>;
    std::mutex mMutex;// mutex for updating mRegister and mFree
    HashTableT mRegister;
    uint8_t*   mData;
    uint8_t*   mFree;
    uint64_t   mSize;
    bool       mManaged;

    /// @brief External memory ctor
    Pool(uint64_t size = 0, void *data = nullptr) : mData(static_cast<uint8_t*>(data)), mFree(mData), mSize(size), mManaged(data==nullptr)
    {
        if (mManaged) {
            // Make sure size is a multiple of alignment
            const uint64_t allocSize = size + NANOVDB_DATA_ALIGNMENT - (size % NANOVDB_DATA_ALIGNMENT);
#if defined(_MSC_VER)
            mData = mFree = static_cast<uint8_t*>(_aligned_malloc(allocSize, NANOVDB_DATA_ALIGNMENT));
#elif defined(__APPLE__)
            posix_memalign(reinterpret_cast<void**>(&mData), NANOVDB_DATA_ALIGNMENT, allocSize);
            mFree = mData;
#else
            mData = mFree = static_cast<uint8_t*>(aligned_alloc(NANOVDB_DATA_ALIGNMENT, allocSize));
#endif
            if (mData == nullptr) {
                throw std::runtime_error("Pool::Pool malloc failed");
            }
        }
    }

    /// @brief Custom destructor
    ~Pool()
    {
        assert(mRegister.empty());
        if (mManaged) {
#if defined(_MSC_VER)
            _aligned_free(mData);
#else
            std::free(mData);
#endif
        }
    }

    /// @brief Disallow copy-construction
    Pool(const Pool&) = delete;

    /// @brief Disallow move-construction
    Pool(const Pool&&) = delete;

    /// @brief Disallow copy assignment operation
    Pool& operator=(const Pool&) = delete;

    /// @brief Disallow move assignment operation
    Pool& operator=(const Pool&&) = delete;

    /// @brief Return the total number of bytes used from this Pool by buffers
    uint64_t usage() const { return static_cast<uint64_t>(mFree - mData); }

    /// @brief Allocate a buffer of the specified size and add it to the register
    void add(AlignedHostBuffer *buffer, uint64_t size)
    {
        if (mFree + size > mData + mSize) {
            std::stringstream ss;
            ss << "AlignedHostBuffer::Pool: insufficient memory\n"
               << "\tA buffer requested " << size << " bytes from a pool with "
               << mSize <<" bytes of which\n\t" << (mFree-mData)
               << " bytes are used by " << mRegister.size() << " other buffer(s). "
               << "Pool is " << (mManaged ? "internally" : "externally") << " managed.\n";
            //std::cerr << ss.str();
            throw std::runtime_error(ss.str());
        }
        buffer->mSize = size;
        const std::lock_guard<std::mutex> lock(mMutex);
        mRegister.insert(buffer);
        buffer->mData = mFree;
        mFree += size;
    }

    /// @brief Remove the specified buffer from the register
    void remove(AlignedHostBuffer *buffer)
    {
        const std::lock_guard<std::mutex> lock(mMutex);
        mRegister.erase(buffer);
    }

    /// @brief Replaces buffer1 with buffer2 in the register
    void replace(AlignedHostBuffer *buffer1, AlignedHostBuffer *buffer2)
    {
        const std::lock_guard<std::mutex> lock(mMutex);
        mRegister.erase( buffer1);
        mRegister.insert(buffer2);
    }

    /// @brief Reset the register and all its buffers
    void reset()
    {
        for (AlignedHostBuffer *buffer : mRegister) {
            buffer->mPool.reset();
            buffer->mSize = 0;
            buffer->mData = nullptr;
        }
        mRegister.clear();
        mFree = mData;
    }
    /// @brief Return true is all the memory in this pool is in use.
    bool isFull() const
    {
        assert(mFree <= mData + mSize);
        return mSize > 0 ? mFree == mData + mSize : false;
    }
};// struct AlignedHostBuffer::Pool

// --------------------------> Implementation of AlignedHostBuffer <------------------------------------

inline AlignedHostBuffer::AlignedHostBuffer(uint64_t size) : mPool(nullptr), mSize(size), mData(nullptr)
{
    if (size>0) {
        mPool = std::make_shared<Pool>(size);
        mData = mPool->mFree;
        mPool->mRegister.insert(this);
        mPool->mFree += size;
    }
}

inline AlignedHostBuffer::AlignedHostBuffer(AlignedHostBuffer&& other) : mPool(other.mPool), mSize(other.mSize), mData(other.mData)
{
    if (mPool && mSize != 0) {
        mPool->replace(&other, this);
    }
    other.mPool.reset();
    other.mSize = 0;
    other.mData = nullptr;
}

inline void AlignedHostBuffer::init(uint64_t bufferSize, void *data)
{
    if (bufferSize == 0) {
        throw std::runtime_error("AlignedHostBuffer: invalid buffer size");
    }
    if (mPool) {
        mPool.reset();
    }
    if (!mPool || mPool->mSize != bufferSize) {
        mPool = std::make_shared<Pool>(bufferSize, data);
    }
    mPool->add(this, bufferSize);
}

inline AlignedHostBuffer& AlignedHostBuffer::operator=(AlignedHostBuffer&& other)
{
    if (mPool) {
        mPool->remove(this);
    }
    mPool = other.mPool;
    mSize = other.mSize;
    mData = other.mData;
    if (mPool && mSize != 0) {
        mPool->replace(&other, this);
    }
    other.mPool.reset();
    other.mSize = 0;
    other.mData = nullptr;
    return *this;
}

inline uint64_t AlignedHostBuffer::poolSize() const
{
    return mPool ? mPool->mSize : 0u;
}

inline uint64_t AlignedHostBuffer::poolUsage() const
{
    return mPool ? mPool->usage(): 0u;
}

inline bool AlignedHostBuffer::isManaged() const
{
    return mPool ? mPool->mManaged : false;
}

inline bool AlignedHostBuffer::isFull() const
{
    return mPool ? mPool->isFull() : false;
}

inline AlignedHostBuffer AlignedHostBuffer::createPool(uint64_t poolSize, void *data)
{
    if (poolSize == 0) {
        throw std::runtime_error("AlignedHostBuffer: invalid pool size");
    }
    AlignedHostBuffer buffer;
    buffer.mPool = std::make_shared<Pool>(poolSize, data);
    // note the buffer is NOT registered by its pool since it is not using its memory
    buffer.mSize = 0;
    buffer.mData = nullptr;
    return buffer;
}

inline AlignedHostBuffer AlignedHostBuffer::createFull(uint64_t bufferSize, void *data)
{
    if (bufferSize == 0) {
        throw std::runtime_error("AlignedHostBuffer: invalid buffer size");
    }
    AlignedHostBuffer buffer;
    buffer.mPool = std::make_shared<Pool>(bufferSize, data);
    buffer.mPool->add(&buffer, bufferSize);
    return buffer;
}

inline AlignedHostBuffer AlignedHostBuffer::create(uint64_t bufferSize, const AlignedHostBuffer* pool)
{
    AlignedHostBuffer buffer;
    if (pool == nullptr || !pool->mPool) {
        buffer.mPool = std::make_shared<Pool>(bufferSize);
    } else {
       buffer.mPool = pool->mPool;
    }
    buffer.mPool->add(&buffer, bufferSize);
    return buffer;
}

inline void AlignedHostBuffer::clear()
{
    if (mPool) {// remove self from the buffer register in the pool
        mPool->remove(this);
    }
    mPool.reset();
    mSize = 0;
    mData = nullptr;
}

inline void AlignedHostBuffer::reset()
{
    if (this->size()>0) {
        throw std::runtime_error("AlignedHostBuffer: only empty buffers can call reset");
    }
    if (!mPool) {
        throw std::runtime_error("AlignedHostBuffer: this buffer contains no pool to reset");
    }
    mPool->reset();
}

} // namespace nanovtt

namespace nanovdb {
    template<>
    struct BufferTraits<nanovtt::AlignedHostBuffer>
    {
        // Set to true to facilitate compilation.
        // Device capabilities detected at runtime instead.
        static const bool hasDeviceDual = true; // false;
    };
}

#if defined(__clang__)
#pragma clang diagnostic pop
#endif

#endif // end of NANOVTT_ALIGNEDHOSTBUFFER_H_HAS_BEEN_INCLUDED
