// Copyright Contributors to the OpenVDB Project
// SPDX-License-Identifier: MPL-2.0

/*!
    \file CopyableCudaDeviceBuffer.h

    \authors Ken Museth and Autodesk

    \brief CopyableCudaDeviceBuffer.h contains modified code from CudaDeviceBuffer.h
           and is copyable which is required for Amino::Ptr.
*/

#ifndef NANOVTT_COPYABLE_CUDA_DEVICE_BUFFER_H_HAS_BEEN_INCLUDED
#define NANOVTT_COPYABLE_CUDA_DEVICE_BUFFER_H_HAS_BEEN_INCLUDED

#if defined(__clang__)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wunused-parameter"
#pragma clang diagnostic ignored "-Wextra-semi"
#pragma clang diagnostic ignored "-Wextra-semi-stmt"
#pragma clang diagnostic ignored "-Wold-style-cast"
#pragma clang diagnostic ignored "-Wimplicit-float-conversion"
#pragma clang diagnostic ignored "-Wmissing-braces"
#pragma clang diagnostic ignored "-Wimplicit-int-conversion"
#pragma clang diagnostic ignored "-Wreserved-id-macro"
#pragma clang diagnostic ignored "-Wcast-qual"
#pragma clang diagnostic ignored "-Wdouble-promotion"
#pragma clang diagnostic ignored "-Wnewline-eof"
#pragma clang diagnostic ignored "-Wshadow"
#pragma clang diagnostic ignored "-Wshorten-64-to-32"
#pragma clang diagnostic ignored "-Wweak-vtables"
#pragma clang diagnostic ignored "-Wunused-template"
#if defined(__clang_major__) && __clang_major__ >= 14
#pragma clang diagnostic ignored "-Wnull-pointer-subtraction"
#endif // defined(__clang_major__) && __clang_major__ >= 14
#endif // defined(__clang__)
#include <Bifrost/GeometryVolume/internal/nanovdb/util/cuda/CudaDeviceBuffer.h>
#if defined(__clang__)
#pragma clang diagnostic pop
#endif // defined(__clang__)
#include <cstring> // memcpy

// Convenience function for checking CUDA runtime API results
// can be wrapped around any runtime API call. No-op in release builds.
#undef cudaCheck
#define cudaCheck(ans) gpuAssert((ans), __FILE__, __LINE__)

#undef checkPtr
#define checkPtr(ptr, msg) ptrAssert((ptr), (msg), __FILE__, __LINE__)

namespace nanovtt {

// ----------------------------> CopyableCudaDeviceBuffer <--------------------------------------

/// @brief Simple memory buffer using un-managed pinned host memory when compiled with NVCC.
///        Obviously this class is making explicit used of CUDA so replace it with your own memory
///        allocator if you are not using CUDA.
/// @note  While CUDA's pinned host memory allows for asynchronous memory copy between host and device
///        it is significantly slower then cached (un-pinned) memory on the host.
class CopyableCudaDeviceBuffer
{
    static constexpr bool mVerbose = false;

    uint64_t mSize; // total number of bytes for the NanoVTT grid.
    uint8_t *mCpuData, *mGpuData; // raw buffer for the NanoVTT grid.


#if defined(DEBUG) || defined(_DEBUG)
    static inline bool gpuAssert(cudaError_t code, const char* file, int line, bool abort = true)
    {
        if (code != cudaSuccess) {
            fprintf(stderr, "CUDA Runtime Error: %s %s %d\n", cudaGetErrorString(code), file, line);
            if (abort)
                exit(code);
            return false;
        }
        return true;
    }
#else
    static inline bool gpuAssert(cudaError_t, const char*, int, bool = true)
    {
        return true;
    }
#endif

#if defined(DEBUG) || defined(_DEBUG)
    static inline void ptrAssert(void* ptr, const char* msg, const char* file, int line, bool abort = true)
    {
        if (ptr == nullptr) {
            fprintf(stderr, "NULL pointer error: %s %s %d\n", msg, file, line);
            if (abort)
                exit(1);
        }
    }
#else
    static inline void ptrAssert(void*, const char*, const char*, int, bool = true)
    {
    }
#endif

public:
    CopyableCudaDeviceBuffer(uint64_t size = 0)
        : mSize(0)
        , mCpuData(nullptr)
        , mGpuData(nullptr)
    {
        this->init(size);
        if (mVerbose)
            std::cout << "CopyableCudaDeviceBuffer.construct(cpu = " << reinterpret_cast<unsigned long long>(mCpuData) << ", gpu = " << reinterpret_cast<unsigned long long>(mGpuData) << ")" << std::endl;
    }
    /// @brief Copy-construction
    CopyableCudaDeviceBuffer(const CopyableCudaDeviceBuffer& other)
    {
        mSize = 0;
        mCpuData = nullptr;
        mGpuData = nullptr;
        if (init(other.mSize)) {
            assert(mSize == other.mSize);
            std::memcpy(mCpuData, other.mCpuData, other.mSize);
            if (other.mGpuData != nullptr)
                deviceUpload();
        }
        if (mVerbose)
            std::cout << "CopyableCudaDeviceBuffer.copy-construct(cpu = " << reinterpret_cast<unsigned long long>(mCpuData) << ", gpu = " << reinterpret_cast<unsigned long long>(mGpuData) << ")" << std::endl;
    }
    /// @brief Move copy-constructors
    CopyableCudaDeviceBuffer(CopyableCudaDeviceBuffer&& other) noexcept
        : mSize(other.mSize)
        , mCpuData(other.mCpuData)
        , mGpuData(other.mGpuData)
    {
        other.mSize = 0;
        other.mCpuData = nullptr;
        other.mGpuData = nullptr;
    }
    /// @brief Copy assignment operation
    CopyableCudaDeviceBuffer& operator=(const CopyableCudaDeviceBuffer& other) noexcept
    {
        clear();
        if (init(other.mSize)) {
            assert(mSize == other.mSize);
            std::memcpy(mCpuData, other.mCpuData, other.mSize);
            if (other.mGpuData != nullptr)
                deviceUpload();
        }
        if (mVerbose)
            std::cout << "CopyableCudaDeviceBuffer.copy(cpu = " << reinterpret_cast<unsigned long long>(mCpuData) << ", gpu = " << reinterpret_cast<unsigned long long>(mGpuData) << ")" << std::endl;
        return *this;
    }
    /// @brief Move copy assignment operation
    CopyableCudaDeviceBuffer& operator=(CopyableCudaDeviceBuffer&& other) noexcept
    {
        clear();
        mSize = other.mSize;
        mCpuData = other.mCpuData;
        mGpuData = other.mGpuData;
        other.mSize = 0;
        other.mCpuData = nullptr;
        other.mGpuData = nullptr;
        return *this;
    }
    /// @brief Destructor frees memory on both the host and device
    ~CopyableCudaDeviceBuffer()
    {
        this->clear();
    }

    bool init(uint64_t size);

    // @brief Retuns a pointer to the raw memory buffer managed by this allocator.
    ///
    /// @warning Note that the pointer can be NULL is the allocator was not initialized!
    uint8_t* data() const { return mCpuData; }
    uint8_t* deviceData() const { return mGpuData; }

    /// @brief Copy grid from the CPU/host to the GPU/device. If @c sync is false the memory copy is asynchronous!
    ///
    /// @note This will allocate memory on the GPU/device if it is not already allocated
    void deviceUpload(void* stream = 0, bool sync = true);

    /// @brief Copy grid from the GPU/device to the CPU/host. If @c sync is false the memory copy is asynchronous!
    void deviceDownload(void* stream = 0, bool sync = true);

    /// @brief Returns the size in bytes of the raw memory buffer managed by this allocator.
    uint64_t size() const { return mSize; }

    /// @brief Returns true if this allocator is empty, i.e. has no allocated memory
    bool empty() const { return mSize == 0; }

    /// @brief De-allocate all memory managed by this allocator and set all pointer to NULL
    void clear();

    /// @brief De-allocate all CPU memory managed by this allocator and set all pointer to NULL
    void clearCPUData();

    /// @brief De-allocate all GPU memory managed by this allocator and set all pointer to NULL
    void clearGPUData();

    static CopyableCudaDeviceBuffer create(uint64_t size, const CopyableCudaDeviceBuffer* context = nullptr);

}; // CopyableCudaDeviceBuffer class

// --------------------------> Implementations below <------------------------------------

inline CopyableCudaDeviceBuffer CopyableCudaDeviceBuffer::create(uint64_t size, const CopyableCudaDeviceBuffer*)
{
    return CopyableCudaDeviceBuffer(size);
}

inline bool CopyableCudaDeviceBuffer::init(uint64_t size)
{
    if (size == mSize)
        return true;
    if (mSize > 0)
        this->clear();
    if (size == 0)
        return true;
    mSize = size;
    cudaCheck(cudaMallocHost(reinterpret_cast<void**>(&mCpuData), size)); // un-managed pinned memory on the host (can be slow to access!)
    checkPtr(mCpuData, "failed to allocate host data");
    return mCpuData != nullptr;
} // CopyableCudaDeviceBuffer::init

inline void CopyableCudaDeviceBuffer::deviceUpload(void* stream, bool sync)
{
    checkPtr(mCpuData, "uninitialized cpu data");
    if (mGpuData == nullptr)
        cudaCheck(cudaMalloc(reinterpret_cast<void**>(&mGpuData), mSize)); // un-managed memory on the device
    checkPtr(mGpuData, "uninitialized gpu data");
    cudaCheck(cudaMemcpyAsync(mGpuData, mCpuData, mSize, cudaMemcpyHostToDevice, reinterpret_cast<cudaStream_t>(stream)));
    if (sync)
        cudaCheck(cudaStreamSynchronize(reinterpret_cast<cudaStream_t>(stream)));
} // CopyableCudaDeviceBuffer::gpuUpload

inline void CopyableCudaDeviceBuffer::deviceDownload(void* stream, bool sync)
{
    checkPtr(mCpuData, "uninitialized cpu data");
    checkPtr(mGpuData, "uninitialized gpu data");
    cudaCheck(cudaMemcpyAsync(mCpuData, mGpuData, mSize, cudaMemcpyDeviceToHost, reinterpret_cast<cudaStream_t>(stream)));
    if (sync)
        cudaCheck(cudaStreamSynchronize(reinterpret_cast<cudaStream_t>(stream)));
} // CopyableCudaDeviceBuffer::gpuDownload

inline void CopyableCudaDeviceBuffer::clear()
{
    clearGPUData();
    clearCPUData();
} // CopyableCudaDeviceBuffer::clear

inline void CopyableCudaDeviceBuffer::clearCPUData()
{
    if (mVerbose)
        std::cout << "CopyableCudaDeviceBuffer.clearCPUData(cpu = " << reinterpret_cast<unsigned long long>(mCpuData) << ", gpu = " << reinterpret_cast<unsigned long long>(mGpuData) << ")" << std::endl;
    if (mCpuData)
        cudaCheck(cudaFreeHost(mCpuData));
    mCpuData = nullptr;
    if (mCpuData == nullptr && mGpuData == nullptr)
        mSize = 0;
} // CopyableCudaDeviceBuffer::clearCPUData

inline void CopyableCudaDeviceBuffer::clearGPUData()
{
    if (mVerbose)
        std::cout << "CopyableCudaDeviceBuffer.clearGPUData(cpu = " << reinterpret_cast<unsigned long long>(mCpuData) << ", gpu = " << reinterpret_cast<unsigned long long>(mGpuData) << ")" << std::endl;
    if (mGpuData)
        cudaCheck(cudaFree(mGpuData));
    mGpuData = nullptr;
    if (mCpuData == nullptr && mGpuData == nullptr)
        mSize = 0;
} // CopyableCudaDeviceBuffer::clearGPUData


} // namespace nanovtt

namespace nanovdb {

template<>
struct BufferTraits<nanovtt::CopyableCudaDeviceBuffer>
{
    static const bool hasDeviceDual = true;
};

} // namespace nanovdb

#endif // end of NANOVTT_CUDA_DEVICE_BUFFER_H_HAS_BEEN_INCLUDED
