// Copyright Contributors to the OpenVDB Project
// SPDX-License-Identifier: MPL-2.0

/*!
    \file GridChecksum.h

    \authors Ken Museth and Autodesk
*/

#ifndef NANOVTT_GRIDCHECKSUM_H_HAS_BEEN_INCLUDED
#define NANOVTT_GRIDCHECKSUM_H_HAS_BEEN_INCLUDED

#if defined(__clang__)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wunused-parameter"
#pragma clang diagnostic ignored "-Wextra-semi"
#pragma clang diagnostic ignored "-Wold-style-cast"
#pragma clang diagnostic ignored "-Wimplicit-float-conversion"
#pragma clang diagnostic ignored "-Wmissing-braces"
#pragma clang diagnostic ignored "-Wimplicit-int-conversion"
#pragma clang diagnostic ignored "-Wreserved-id-macro"
#pragma clang diagnostic ignored "-Wcast-qual"
#pragma clang diagnostic ignored "-Wdouble-promotion"
#pragma clang diagnostic ignored "-Wnewline-eof"
#pragma clang diagnostic ignored "-Wshadow"
#pragma clang diagnostic ignored "-Wshorten-64-to-32"
#pragma clang diagnostic ignored "-Wextra-semi-stmt"
#endif
#include <Bifrost/GeometryVolume/internal/nanovdb/util/GridChecksum.h>
#if defined(__clang__)
#pragma clang diagnostic pop
#endif

namespace nanovtt {

/// @brief Class that computes the Cyclic Redundancy Check (CRC)
class CRC32 {
    using ByteT = std::uint_fast8_t;
    using HashT = std::uint_fast32_t;
    HashT mChecksum;

    static std::array<HashT, 256> INIT() {
        HashT n      = 0;
        auto  kernel = [&n]() -> HashT {
            HashT checksum = n++;
            for (int i = 0; i < 8; ++i)
                checksum = (checksum >> 1) ^ ((checksum & 0x1u) ? HashT{0xEDB88320uL} : 0);
            return checksum;
        };
        std::array<HashT, 256> LUT{};
        std::generate(LUT.begin(), LUT.end(), kernel);
        return LUT; // move semantic should prevent a deep copy
    }

public:
    static const HashT EMPTY = ~HashT{0} & HashT{0xFFFFFFFFuL}; // All bits are on

    CRC32() : mChecksum(EMPTY) {}

    void reset() { mChecksum = EMPTY; }

    HashT checksum() const { return HashT{0xFFFFFFFFuL} & ~mChecksum; }

    template <typename IterT>
    void operator()(IterT begin, IterT end) {
        static const auto LUT = INIT(); // scoped static initialization is thread-safe since C++11
        auto              kernel = [](HashT checksum, ByteT value) {
            return LUT[(checksum ^ value) & 0xFFu] ^ (checksum >> 8);
        };
        mChecksum = std::accumulate(begin, end, mChecksum, kernel);
    }

    void operator()(const void* data, size_t byteSize) {
        const ByteT* begin = static_cast<const ByteT*>(data);
        this->       operator()<const ByteT*>(begin, begin + byteSize);
    }
}; // CRC32


/// @brief Return the (2 x CRC32) checksum of the specified @a grid
///
/// @param grid Grid from which the checksum is computed.
/// @param mode Defines the mode of computation for the checksum.
template <typename ValueT>
uint64_t checksum(const NanoVTTGrid<ValueT> &grid, ::nanovdb::ChecksumMode mode = ::nanovdb::ChecksumMode::Default);

/// @brief Return true if the checksum of the @a grid matches the expected
///        value already encoded into the grid's meta data.
///
/// @param grid Grid whose checksum is validated.
/// @param mode Defines the mode of computation for the checksum.
template <typename ValueT>
bool validateChecksum(const NanoVTTGrid<ValueT> &grid, ::nanovdb::ChecksumMode mode = ::nanovdb::ChecksumMode::Default);

/// @brief Updates the checksum of a grid
///
/// @param grid Grid whose checksum will be updated.
/// @param mode Defines the mode of computation for the checksum.
template <typename ValueT>
void updateChecksum(NanoVTTGrid<ValueT> &grid, ::nanovdb::ChecksumMode mode = ::nanovdb::ChecksumMode::Default);

template <typename ValueT>
nanovdb::GridChecksum gridChecksum(const NanoVTTGrid<ValueT> &grid, ::nanovdb::ChecksumMode mode)
{
    static const size_t offset = 16;

    if (mode == ::nanovdb::ChecksumMode::Disable) return nanovdb::GridChecksum();

    const auto &tree = grid.tree();
    nanovtt::CRC32 crc;
    uint8_t const* begin;
    uint8_t const* end;
    uint32_t mCRC[2] = { nanovtt::CRC32::EMPTY, nanovtt::CRC32::EMPTY };

    // process Container + (Grid + TreeBase), but exclude mMagic and mChecksum
    std::vector<std::uint_fast32_t> checksums;
    begin = reinterpret_cast<const uint8_t*>(tree.getContainer());
    end = begin + tree.getContainer()->memUsage();
    crc(begin, end);
    checksums.push_back(crc.checksum());
    crc.reset();
    begin = reinterpret_cast<const uint8_t*>(&grid);
    end = begin + grid.memUsage() + sizeof(nanovtt::TreeBase<ValueT>);
    crc(begin + offset, end);
    checksums.push_back(crc.checksum());
    crc.reset();
    crc(checksums.data(), sizeof(std::uint_fast32_t) * checksums.size());
    mCRC[0] = static_cast<uint32_t>(crc.checksum());
    crc.reset();

    if (mode == ::nanovdb::ChecksumMode::Partial || tree.isEmpty()) return nanovdb::GridChecksum(mCRC[0], mCRC[1]);

    // process Tree Data
    begin = reinterpret_cast<const uint8_t*>(&tree);
    end = begin + tree.memUsage();
    crc(begin + sizeof(nanovtt::TreeBase<ValueT>), end);
    mCRC[1] = static_cast<uint32_t>(crc.checksum());

    return nanovdb::GridChecksum(mCRC[0], mCRC[1]);
}

template <typename ValueT>
uint64_t checksum(const NanoVTTGrid<ValueT> &grid, ::nanovdb::ChecksumMode mode)
{
    nanovdb::GridChecksum cs = gridChecksum(grid, mode);
    return cs.checksum();
}

template <typename ValueT>
bool validateChecksum(const NanoVTTGrid<ValueT> &grid, ::nanovdb::ChecksumMode mode)
{
    nanovdb::GridChecksum cs1(grid.checksum(), mode), cs2;
    cs2 = gridChecksum(grid, cs1.mode());
    return cs1 == cs2;
}

template <typename ValueT>
void updateChecksum(NanoVTTGrid<ValueT> &grid, ::nanovdb::ChecksumMode mode)
{
    nanovdb::GridChecksum cs = gridChecksum(grid, mode);
    grid.data()->mChecksum = cs.checksum();
}

} // namespace nanovtt

#endif // NANOVTT_GRIDCHECKSUM_H_HAS_BEEN_INCLUDED
