// Copyright Contributors to the OpenVDB Project
// SPDX-License-Identifier: MPL-2.0

/*!
    \file VTTToNanoVTT.h

    \authors Ken Museth and Autodesk

    \brief VTTToNanoVTT.h contains modified code from OpenToNanoVDB.h.
*/

#ifndef NANOVTT_VTTTONANOVTT_H_HAS_BEEN_INCLUDED
#define NANOVTT_VTTTONANOVTT_H_HAS_BEEN_INCLUDED

#include <Bifrost/Math/tuple_helpers.h>
#include <Bifrost/Math/bbox.h>
#include <Bifrost/Math/Types.h>
#include <Bifrost/GeometryVolume/object.h>
#include <Bifrost/GeometryVolume/nanovtt/util/GridHandle.h>
#include <Bifrost/GeometryVolume/nanovtt/util/Quantization.h>
#if defined(__clang__)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wunused-parameter"
#pragma clang diagnostic ignored "-Wextra-semi"
#pragma clang diagnostic ignored "-Wold-style-cast"
#pragma clang diagnostic ignored "-Wimplicit-float-conversion"
#pragma clang diagnostic ignored "-Wmissing-braces"
#pragma clang diagnostic ignored "-Wimplicit-int-conversion"
#pragma clang diagnostic ignored "-Wreserved-id-macro"
#pragma clang diagnostic ignored "-Wcast-qual"
#pragma clang diagnostic ignored "-Wdouble-promotion"
#pragma clang diagnostic ignored "-Wnewline-eof"
#pragma clang diagnostic ignored "-Wshadow"
#pragma clang diagnostic ignored "-Wshorten-64-to-32"
#pragma clang diagnostic ignored "-Wweak-vtables"
#endif
#include <Bifrost/GeometryVolume/internal/nanovdb/util/HostBuffer.h>
#include <Bifrost/GeometryVolume/internal/nanovdb/util/GridChecksum.h>
#include <Bifrost/GeometryVolume/internal/nanovdb/util/GridStats.h>
#if defined(__clang__)
#pragma clang diagnostic pop
#endif

namespace nanovtt
{

/// \brief Map from VTT type to NanoVTT type
/// Conversion from VTT to NanoVTT supports the same types as NanoVDB except:
/// - ValueMask
/// - bool
/// - Rgba8
/// - Vec4d
/// - \todo BIFROST-TBD Clang complains about Amino::tuple<Amino::long_t, int64_t>. I suspect it's a bug in the compiler.
using NanoVTTTypes = Amino::tuple<
    Amino::tuple<Amino::float_t, float>,
    Amino::tuple<Amino::double_t, double>,
    Amino::tuple<Amino::short_t, int16_t>,
    Amino::tuple<Amino::int_t, int32_t>,
#if !defined(__clang__)
	Amino::tuple<Amino::long_t, int64_t>,
#endif
    Amino::tuple<Bifrost::Math::float3, nanovdb::Vec3f>,
    Amino::tuple<Bifrost::Math::double3, nanovdb::Vec3d>,
    Amino::tuple<Amino::uint_t, uint32_t>,
    Amino::tuple<Bifrost::Math::float4, nanovdb::Vec4f>
>;

/// <summary>
///  Converts a VTT with Traits24 to a NanoVTT.
/// </summary>
/// <typeparam name="BufferT">The buffer type. Supports nanovdb::HostBuffer and nanovtt::CopyableCudaDeviceBuffer.</typeparam>
/// <typeparam name="VoxelT">The voxel quantization type for float. Supports void (no quantization), nanovdb::Fp4, nanovdb::Fp8, nanovdb::Fp16, nanovdb::FpN.</typeparam>
/// <typeparam name="OracleT">The method used for FpN quantization.</typeparam>
/// <typeparam name="kEnableCoalescedTilesSupportT">Enables conversion support for coalescing tiles into larger tiles. Must be used in conjunction with NANOVTT_SUPPORT_COALESCED_TILES defined.</typeparam>
/// <param name="in_volume">The VTT volume.</param>
/// <param name="in_properties">The properties that should be converted.</param>
/// <param name="in_sMode">The mode for computing stats.</param>
/// <param name="in_cMode">The mode for computing the checksum.</param>
/// <param name="in_verbose">The verbosity level.</param>
/// <param name="in_tolerance">The tolerance used in the oracle for quantizing values.</param>
/// <param name="in_ditherOn">True if dithering is enabled for quantized values.</param>
/// <param name="in_extendedBudgetFraction">The fraction (between zero and one) of additional space the extended tile for sampling is allowed to take up. Larger fractions take more space but allow faster linear sampling.</param>
/// <param name="in_minAdaptiveLevels">The minimum number of levels that are represented adaptively. A smaller number leads to a shallower tree and potentially faster random access.</param>
/// <param name="in_discardFinestLevels">The number of finest levels that are discarded (not converted). Can be used to reduce memory requirements and increase performance, at the cost of quality.</param>
/// <param name="in_sortIntoMortonOrder">If true the tiles are sorted into Morton order. Coversion will be slower, but accessing values will be faster.</param>
/// <param name="in_coalesceTileData">If true, tiles are coalesced into larger tiles if possible. In some cases this leads to a smaller memory footprint when using extended tiles.</param>
/// <param name="in_padBufferForVolumetricTexture">Will ensure that the size of the output buffer is of size in_textureElementSize * N^3 for some N. This means the buffer can be used directly as a volumetric texture without any additional copying.</param>
/// <param name="in_textureElementSize">The size of a texture element.</param>
/// <param name="in_storeTileDataAs3DTextures">Will store all tile data in a 3D textures format for use e.g. in a shader.</param>
/// <param name="in_dilateActiveFogVoxels">If enabled will dilate the active fog voxels to ensure that raymarching using HDDA will work.</param>
/// <param name="in_activeFogVoxelsThreshold">The threshold used to determine active fog voxels.</param>
/// <param name="in_setCoarserLevels">Ensures that coarser levels in the NanoVTT are set to facilitate LOD rendering.</param>
/// <param name="in_enableNanoVTTTraversal">Adds a structure to the NanoVTT which facilitates parallel traversal over all voxels with access to tile coordinates.</param>
/// <param name="out_traversalStructure">The traversal structure will be added to this buffer if not nullptr.</param>
/// <param name="out_neighborStructure">The neighbor structure will be added to this buffer if not nullptr.</param>
/// <param name="out_bbox">A tight-fitting bounding box. May not be identical to the one stored in the output NanoVTT in case in_dilateActiveFogVoxels is true.</param>
/// <param name="in_allocator">An instantiation of the buffer used to allocate memory for the NanoVTT.</param>
/// <returns>The GridHandle from which the NanoVTT can be retrieved.</returns>
template <typename BufferT = nanovdb::HostBuffer, typename VoxelT = void, typename OracleT = nanovdb::AbsDiff, bool kEnableCoalescedTilesSupportT = false>
GridHandle<BufferT> vttToNanoVTT(
    const BIF::Object& in_volume,
    const Amino::String& in_properties,
    const nanovdb::StatsMode    in_sMode = nanovdb::StatsMode::Default,
    const nanovdb::ChecksumMode in_cMode = nanovdb::ChecksumMode::Default,
    const int                   in_verbose = 0,
    const float                 in_tolerance = -1.f,
    const bool                  in_ditherOn = false,
    const float                 in_extendedBudgetFraction = 0.2f,
    const int                   in_minAdaptiveLevels = 4,
    const int                   in_discardFinestLevels = 0,
    const bool                  in_sortIntoMortonOrder = false,
    const bool                  in_coalesceTileData = false,
    const bool                  in_padBufferForVolumetricTexture = false,
    const int                   in_textureElementSize = 4,
    const bool                  in_storeTileDataAs3DTextures = false,
    const bool                  in_dilateActiveFogVoxels = false,
    const float                 in_activeFogVoxelsThreshold = 0.001f,
    const bool                  in_setCoarserLevels = false,
    const bool                  in_enableNanoVTTTraversal = false,
    BufferT*                    out_traversalStructure = nullptr,
    BufferT*                    out_neighborStructure = nullptr,
    Bifrost::Math::bboxf*       out_bbox = nullptr,
    const BufferT&              in_allocator = BufferT());


/// @brief This class will convert a VTT grid into a NanoVTT grid managed by a GridHandle.
template<typename NanoBuildT = void,
    typename OracleT = nanovdb::AbsDiff,
    typename BufferT = nanovdb::HostBuffer>
class VTTToNanoVTT
{
    int                   mVerbose;
    nanovdb::StatsMode    mStats;
    nanovdb::ChecksumMode mChecksum;
    bool                  mDitherOn;
    float                 mExtendedBudgetFraction;
    int                   mMinAdaptiveLevels;
    int                   mDiscardFinestLevels;
    bool                  mSortIntoMortonOrder;
    bool                  mCoalesceTileData;
    OracleT               mOracle;
    Amino::String         mProperties;
    bool                  mPadBufferForVolumetricTexture;
    int                   mTextureElementSize;
    bool                  mStoreTileDataAs3DTextures;
    bool                  mDilateActiveFogVoxels;
    float                 mActiveFogVoxelsThreshold;
    bool                  mSetCoarserLevels;
    bool                  mEnableNanoVTTTraversal;
    BufferT*              mTraversalStructure;
    BufferT*              mNeighborStructure;
    Bifrost::Math::bboxf* mBBox;

public:
    /// @brief Default c-tor
    VTTToNanoVTT()
        : mVerbose(0)
        , mStats(nanovdb::StatsMode::Default)
        , mChecksum(nanovdb::ChecksumMode::Default)
        , mDitherOn(false)
        , mExtendedBudgetFraction(0.2f)
        , mMinAdaptiveLevels(4)
        , mDiscardFinestLevels(0)
        , mSortIntoMortonOrder(false)
        , mCoalesceTileData(false)
        , mOracle(-1.f)
        , mProperties("*")
        , mPadBufferForVolumetricTexture(false)
        , mTextureElementSize(4)
        , mStoreTileDataAs3DTextures(false)
        , mDilateActiveFogVoxels(false)
        , mActiveFogVoxelsThreshold(0.001f)
        , mSetCoarserLevels(false)
        , mEnableNanoVTTTraversal(false)
        , mTraversalStructure(nullptr)
        , mNeighborStructure(nullptr)
        , mBBox(nullptr)
    {
    }

    /// @brief return a reference to the compression oracle
    ///
    /// @note Note, the oracle is only used when NanoBuildT = nanovdb::FpN!
    OracleT& oracle() { return mOracle; }

    void setVerbose(int mode = 1) { mVerbose = mode; }

    void enableDithering(bool on = true) { mDitherOn = on; }

    void setExtendedBudgetFraction(const float extendedBudgetFraction) { mExtendedBudgetFraction = extendedBudgetFraction; }

    void setMinAdaptiveLevels(const int minAdaptiveLevels) { mMinAdaptiveLevels = minAdaptiveLevels; }

    void setDiscardFinestLevels(const int discardFinestLevels) { mDiscardFinestLevels = discardFinestLevels; }

    void setSortIntoMortonOrder(const bool sortIntoMortonOrder) { mSortIntoMortonOrder = sortIntoMortonOrder; }

    void setCoalesceTileData(const bool coalesceTileData) { mCoalesceTileData = coalesceTileData; }

    void setStats(nanovdb::StatsMode mode = nanovdb::StatsMode::Default) { mStats = mode; }

    void setChecksum(nanovdb::ChecksumMode mode = nanovdb::ChecksumMode::Default) { mChecksum = mode; }

    void setProperties(const Amino::String& properties = "*") { mProperties = properties; }

    void setPadBufferForVolumetricTexture(const bool padBufferForVolumetricTexture = false) { mPadBufferForVolumetricTexture = padBufferForVolumetricTexture;  }

    void setVolumetricTextureElementSize(const int textureElementSize = 4) { mTextureElementSize = textureElementSize; }

    void setStoreTileDataAs3DTextures(const bool storeTileDataAs3DTextures = false) { mStoreTileDataAs3DTextures = storeTileDataAs3DTextures; }

    void setDilateActiveFogVoxels(const bool dilateActiveFogVoxels = false) { mDilateActiveFogVoxels = dilateActiveFogVoxels; }

    void setCoarserLevels(const float setCoarserLevels = false) { mSetCoarserLevels = setCoarserLevels; }

    void setActiveFogVoxelsThreshold(const float activeFogVoxelsThreshold = 0.001f) { mActiveFogVoxelsThreshold = activeFogVoxelsThreshold; }

    void setEnableNanoVTTTraversal(const bool enableNanoVTTTraversal) { mEnableNanoVTTTraversal = enableNanoVTTTraversal; }

    void setTraversalStructure(BufferT* traversalStructure) { mTraversalStructure = traversalStructure; }

    void setNeigborStructure(BufferT* neighborStructure) { mNeighborStructure = neighborStructure; }

    void setBBox(Bifrost::Math::bboxf* bbox) { mBBox = bbox; }

    /// @brief Return a shared pointer to a NanoVTT grid handle constructed from
    /// the specified OpenVDB grid
    GridHandle<BufferT> operator()(const BIF::Object& volume,
                                   const BufferT&     allocator = BufferT()) {
        return vttToNanoVTT<BufferT, NanoBuildT, OracleT>(
            volume, mProperties, mStats, mChecksum, mVerbose,
            mOracle.getTolerance(), mDitherOn, mExtendedBudgetFraction,
            mMinAdaptiveLevels, mDiscardFinestLevels, mSortIntoMortonOrder,
            mCoalesceTileData, mPadBufferForVolumetricTexture,
            mTextureElementSize, mStoreTileDataAs3DTextures,
            mDilateActiveFogVoxels, mActiveFogVoxelsThreshold,
            mSetCoarserLevels, mEnableNanoVTTTraversal, mTraversalStructure,
            mNeighborStructure, mBBox, allocator);
    }

    GridHandle<BufferT> operator()(const BIF::Object&    volume,
                                   nanovdb::StatsMode    sMode,
                                   nanovdb::ChecksumMode cMode,
                                   int                   verbose,
                                   const BufferT& allocator = BufferT()) {
        return vttToNanoVTT<BufferT, NanoBuildT, OracleT>(
            volume, mProperties, sMode, cMode, verbose, mOracle.getTolerance(),
            mDitherOn, mExtendedBudgetFraction, mMinAdaptiveLevels,
            mDiscardFinestLevels, mSortIntoMortonOrder, mCoalesceTileData,
            mPadBufferForVolumetricTexture, mTextureElementSize,
            mStoreTileDataAs3DTextures, mDilateActiveFogVoxels,
            mActiveFogVoxelsThreshold, mSetCoarserLevels,
            mEnableNanoVTTTraversal, mTraversalStructure, mNeighborStructure,
            mBBox, allocator);
    }
};


} // namespace nanovtt

#endif // NANOVTT_VTTTONANOVTT_H_HAS_BEEN_INCLUDED
