// Copyright Contributors to the OpenVDB Project
// SPDX-License-Identifier: MPL-2.0

/*!
    \file VTTToNanoVTTImpl.h

    \authors Ken Museth and Autodesk

    \brief Contains modified code from OpenToNanoVDB.h.
*/

#include <cstring>
#include <type_traits>
#include <chrono>
#include <thread>
#include <Bifrost/GeometryVolume/bif_decl.h>
#include <Bifrost/Math/Math_operators.h>
#include <Bifrost/Math/int3_helpers.h>
#include <Bifrost/Math/float2_helpers.h>
#include <Bifrost/Math/float3_helpers.h>
#include <Bifrost/Math/float4_helpers.h>
#include <Bifrost/Math/double2_helpers.h>
#include <Bifrost/Math/double3_helpers.h>
#include <Bifrost/Math/double4_helpers.h>
#include <Bifrost/Math/vector.h>
#include <Bifrost/Math/levelset.h>
#include <Bifrost/GeometryVolume/tile_tree_3d/channel.h>
#include <Bifrost/GeometryVolume/tile_tree_3d/channel_pool.h>
#include <Bifrost/GeometryVolume/tile_tree_3d/channel_set_nonleafvalue_to_childvalue.h>
#include <Bifrost/GeometryVolume/tile_tree_3d/container.h>
#include <Bifrost/GeometryVolume/tile_tree_3d/voxel_topology.h>
#include <Bifrost/GeometryVolume/tile_tree_3d/tile_tree_traits.h>
#include <Bifrost/GeometryVolume/tile_tree_3d/tile_data_halo_helper.h>
#include <Bifrost/GeometryVolume/tile_tree_3d/levelset/is_signed_distance.h>
#include <Bifrost/GeometryVolume/tile_tree_3d/morton.h>
#include <tbb/tick_count.h>
#include <tbb/atomic.h>
#include <tbb/parallel_sort.h>
#include <tbb/parallel_for.h>
#include <Bifrost/GeometryVolume/nanovtt/NanoVTT.h>
#include <Bifrost/GeometryVolume/nanovtt/util/AlignedHostBuffer.h>
#include <Bifrost/GeometryVolume/nanovtt/util/VTTToNanoVTT.h>
#include <Bifrost/GeometryVolume/nanovtt/util/Quantization.h>
#include <Bifrost/GeometryVolume/nanovtt/util/GridChecksum.h>
#include <Bifrost/GeometryVolume/channelparser.h>
#include <Bifrost/GeometryVolume/internal/nanovdb/util/DitherLUT.h>

namespace Internal {

// Clamping SDF voxel values can avoid holes in level set surfaces if default (very large)
// values are next to the zero crossing. Disabled by default.
inline constexpr bool kClampSDFVoxelValues = false;
inline constexpr float kClampSDFVoxelValuesThreshold = 10.f;

// \todo BIFROST-TBD : remove
template <typename T>
inline T* temp_const_cast(T const* t)
{
    return const_cast<T*>(t);
}

// \todo BIFROST-TBD : remove
template <typename T>
inline T& temp_const_cast(T const& t)
{
    return const_cast<T&>(t);
}

inline constexpr size_t k2DTextureRowSize = 16384;

// VTT only supports the Traits<2, 4> configuration
using BIFTraits        = BIF::TileTree3D::Traits<2, 4>;
using BIFContainer     = BIF::TileTree3D::Container<BIFTraits>;
using BIFVoxelTopology = BIF::TileTree3D::VoxelTopology<BIFTraits>;
using BIFTile = BIFContainer::TileType;
using BIFTileAccessor = BIFContainer::ConstTileAccessor;

#ifdef PEPTIDE_DEBUG
template <typename BuildType>
inline void testTileData(const BIFContainer& /*in_container*/,
    nanovtt::Tree<BuildType> const& /*in_tree*/,
    int64_t const* /*in_tileCountAccum*/,
    Amino::Vector<nanovtt::TileIndex> const* /*in_indexToVTT*/,
    const float                     /*in_default*/,
    const int                       /*in_rootDepth*/,
    const int                       /*in_kMaxDepth*/)
{
    // NOP
}

template <>
inline void testTileData(const BIFContainer&             in_container,
                  nanovtt::Tree<float>            const& in_tree,
                  int64_t const*                  in_tileCountAccum,
                  Amino::Vector<nanovtt::TileIndex> const* in_indexToVTT,
                  const float                     in_default,
                  const int                       in_rootDepth,
                  const int                       in_kMaxDepth) {
    static constexpr float epsilon = std::numeric_limits<float>::epsilon();
    in_container.traverse(
        [&](const auto& in_tile) {
            const nanovtt::TileIndex accum =
                static_cast<nanovtt::TileIndex>(in_tileCountAccum[in_tile.depth()]);
            const nanovtt::TileIndex indexVTT =
                accum + in_indexToVTT[in_tile.depth()][in_tile.index()];
            float const* td = in_tree.tileData(indexVTT);

            for (int k = 0; k < BIFTraits::kTileDataWidth; ++k)
                for (int j = 0; j < BIFTraits::kTileDataWidth; ++j)
                    for (int i = 0; i < BIFTraits::kTileDataWidth; ++i) {
                        const int linearIndex =
                            k * (BIFTraits::kTileDataWidth + 1) *
                                (BIFTraits::kTileDataWidth + 1) +
                            j * (BIFTraits::kTileDataWidth + 1) + i;
                        if (td[linearIndex] <= in_default) {
                            bool done = false;
                            for (int kk = std::max(k - 1, 0);
                                 kk <= std::min(k + 1,
                                                BIFTraits::kTileDataWidth) &&
                                 !done;
                                 ++kk)
                                for (int jj = std::max(j - 1, 0);
                                     jj <=
                                         std::min(j + 1,
                                                  BIFTraits::kTileDataWidth) &&
                                     !done;
                                     ++jj)
                                    for (int ii = std::max(i - 1, 0);
                                         ii <= std::min(
                                                   i + 1,
                                                   BIFTraits::kTileDataWidth) &&
                                         !done;
                                         ++ii) {
                                        const int linearIndex2 =
                                            kk *
                                                (BIFTraits::kTileDataWidth +
                                                 1) *
                                                (BIFTraits::kTileDataWidth +
                                                 1) +
                                            jj * (BIFTraits::kTileDataWidth +
                                                  1) +
                                            ii;
                                        assert(td[linearIndex2] <= epsilon + in_default);
                                        if (td[linearIndex2] > epsilon + in_default) {
                                            std::cout
                                                << "ERR: at tile = (" << ii
                                                << ", " << jj << ", " << kk
                                                << ") offset from (" << i
                                                << ", " << j << ", " << k
                                                << ") = " << td[linearIndex2]
                                                << std::endl;
                                            done = true;
                                        }
                                    }
                        }
                    }
        },
        BIF::TileTree3D::TraversalMode::kParallel,
        BIF::TileTree3D::kInvalidTreeIndex, in_rootDepth, in_kMaxDepth);
}
#endif // PEPTIDE_DEBUG

} // namespace

// Note: currently disabled as it does not lead to speedups.
//#define OPTIMIZE_USING_PARENT_HALO

namespace Bifrost
{
namespace Geometry
{
// Note: the converter cannot depend on Bifrost::Geometry so we have hardcoded the names here
static const char* sVoxelComp = "voxel_component";
static const char* sVoxelTileTree = "voxel_tile_tree";
static const char* sVoxelFogDensity = "voxel_fog_density";
static const char* sVoxelSignedDistance = "voxel_signed_distance";
}
}

namespace nanovtt
{
    using namespace Internal;

    template<typename T>
    static bool isSignificant(const bool /*in_isSDF*/, const float /*in_sdfThreshold*/, const T& in_defaultValue, const T& in_val, bool* /*out_isAboveThreshold*/, const float /*in_threshold*/)
    {
        // In the general case we ignore isSDF.
        // We use mag2 to avoid sqrt here.
        return Bifrost::Math::mag2(in_defaultValue - in_val) > 0;
    }

    static bool isSignificant(const bool in_isSDF, const float in_sdfThreshold, const float& in_defaultValue, const float& in_val, bool* out_isAboveThreshold, const float in_threshold)
    {
        if (out_isAboveThreshold) {
            const float absVal = std::abs(in_defaultValue - in_val);
            *out_isAboveThreshold = (absVal > in_threshold + std::numeric_limits<float>::epsilon());
            return (in_isSDF ? Bifrost::Math::isInside(in_val, in_sdfThreshold) : absVal > 0);
        }
        else {
            return (in_isSDF ? Bifrost::Math::isInside(in_val, in_sdfThreshold) : std::abs(in_defaultValue - in_val) > 0);
        }
    }

    template<typename TDH>
    void clampSDFVoxelValues(const float /*in_default*/, TDH& io_tdh, const float in_maxValue)
    {
        for (int k = 0; k <= Traits::kTileDataWidth; ++k) {
            for (int j = 0; j <= Traits::kTileDataWidth; ++j) {
                for (int i = 0; i <= Traits::kTileDataWidth; ++i) {
                    float& val = io_tdh(i, j, k);
                    if (fabsf(val) > in_maxValue) {
                        val = std::copysign(1.f, val) * in_maxValue;
                    }
                }
            }
        } // for (int k = 0; k <= Traits::kTileDataWidth; ++k)
    }

    template<typename VTTType, typename TDH>
    void clampSDFVoxelValues(const VTTType /*in_default*/, TDH& /*io_tdh*/, const float /*in_maxValue*/)
    {
    }

    template<typename VTTType>
    VTTType clampSDFDefaultValue(const VTTType in_default, const float /*in_voxelSize*/)
    {
        return in_default;
    }

    static float clampSDFDefaultValue(const float /*in_default*/, const float in_voxelSize)
    {
        return kClampSDFVoxelValuesThreshold * in_voxelSize;
    }

    template<typename TDH>
    void setActiveState(const float /*in_default*/, TDH& io_tdh1, TDH& io_tdh2, const float in_threshold)
    {
        static constexpr float epsilon = std::numeric_limits<float>::epsilon();
        // Three separable passes instead of one pass with a 27-point stencil.
        // When doing this, this step has almost no additional cost.
        for (int k = 0; k <= Traits::kTileDataWidth; ++k) {
            for (int j = 0; j <= Traits::kTileDataWidth; ++j) {
                for (int i = 0; i <= Traits::kTileDataWidth; ++i) {
                    if (io_tdh1(i, j, k) <= in_threshold &&
                        (io_tdh1(std::max(i - 1, 0), j, k) > in_threshold ||
                            io_tdh1(std::min(i + 1, Traits::kTileDataWidth), j,
                                k) > in_threshold)) {
                        io_tdh2(i, j, k) = epsilon + in_threshold;
                    }
                    else {
                        io_tdh2(i, j, k) = io_tdh1(i, j, k);
                    }
                }
            }
        } // for (int k = 0; k <= Traits::kTileDataWidth; ++k)
        for (int k = 0; k <= Traits::kTileDataWidth; ++k) {
            for (int j = 0; j <= Traits::kTileDataWidth; ++j) {
                for (int i = 0; i <= Traits::kTileDataWidth; ++i) {
                    if (io_tdh2(i, j, k) <= in_threshold &&
                        (io_tdh2(i, std::max(j - 1, 0), k) > in_threshold ||
                         io_tdh2(i, std::min(j + 1, Traits::kTileDataWidth),
                                 k) > in_threshold)) {
                        io_tdh1(i, j, k) = epsilon + in_threshold;
                    } else {
                        io_tdh1(i, j, k) = io_tdh2(i, j, k);
                    }
                }
            }
        } // for (int k = 0; k <= Traits::kTileDataWidth; ++k)
        for (int k = 0; k <= Traits::kTileDataWidth; ++k) {
            for (int j = 0; j <= Traits::kTileDataWidth; ++j) {
                for (int i = 0; i <= Traits::kTileDataWidth; ++i) {
                    if (io_tdh1(i, j, k) <= in_threshold &&
                        (io_tdh1(i, j, std::max(k - 1, 0)) > in_threshold ||
                         io_tdh1(i, j,
                                 std::min(k + 1, Traits::kTileDataWidth)) >
                             in_threshold)) {
                        io_tdh2(i, j, k) = epsilon + in_threshold;
                    } else {
                        io_tdh2(i, j, k) = io_tdh1(i, j, k);
                    }
                }
            }
        } // for (int k = 0; k <= Traits::kTileDataWidth; ++k)

#ifdef PEPTIDE_DEBUG
        for (int k = 0; k < Traits::kTileDataWidth; ++k)
            for (int j = 0; j < Traits::kTileDataWidth; ++j)
                for (int i = 0; i < Traits::kTileDataWidth; ++i) {
                    if (io_tdh2(i, j, k) <= in_threshold) {
                        bool done = false;
                        for (int kk = std::max(k - 1, 0); kk <= std::min(k + 1, Traits::kTileDataWidth) && !done; ++kk)
                            for (int jj = std::max(j - 1, 0); jj <= std::min(j + 1, Traits::kTileDataWidth) && !done; ++jj)
                                for (int ii = std::max(i - 1, 0); ii <= std::min(i + 1, Traits::kTileDataWidth) && !done; ++ii) {
                                    assert(io_tdh2(ii, jj, kk) <= in_threshold + epsilon);
                                }
                    }
                }
#endif

    }

    template<typename VTTType, typename TDH>
    void setActiveState(const VTTType /*in_default*/, TDH& /*io_tdh1*/, TDH& /*io_tdh2*/, const float /*in_threshold*/)
    {
    }

    template <typename VoxelT,
              typename BuildType,
              typename VTTType,
              typename TD,
              typename TDH,
              int valueTypeSize>
    struct Visitor : public BIFContainer::VisitorType {
        Tree<BuildType>* const          m_treeVTT;
        nanovdb::DitherLUT              m_lut;
        TD                              m_tdIn;
        TDH                             m_tdhIn;
        TDH                             m_tdhIn2;
        const Codec                     m_dummyCodec;
        BIFTileAccessor                 m_acc;
        Container const* const          m_containerVTT;
        Amino::Vector<uint64_t> const&  m_tdOffset;
        const uint64_t                  m_tdOffsetConstant;
        Amino::Vector<TileIndex> const* m_indexToVTT;
        Amino::Vector<uint64_t> const&      m_tdcOffset;
        Amino::Vector<int> const*           m_tilesToCoalesce;
        const int                           m_rootDepth;
        const int                           m_maxDepth;
        const bool                          m_coalesceTileData;
        Amino::Vector<unsigned char> const* m_tilesToRemove;
        BIF::TileTree3D::Channel<BIFTraits,
                                 VTTType,
                                 BIFTraits::kTileDataWidth> const& m_channel;
        Amino::Vector<Amino::Vector<Codec>> const&                 m_codecsVec;
        const uint64_t                                             m_fpNIndex;
        Amino::Vector<Amino::Vector<uint64_t>> const& m_tdOffsetFpNVec;
        const bool                                    m_isSDF;
        const float                                   m_voxelSize;
        const bool                                    m_sortIntoMortonOrder;
        const Amino::Vector<int64_t>*                 m_permutation;
        const bool                                    m_dilateActiveFogVoxels;
        const bool                                    m_enableActiveState;
        const float                                   m_activeFogVoxelsThreshold;
        int                                           m_imin;
        int                                           m_jmin;
        int                                           m_kmin;
        int                                           m_imax;
        int                                           m_jmax;
        int                                           m_kmax;
        int                                           m_imintf;
        int                                           m_jmintf;
        int                                           m_kmintf;
        int                                           m_imaxtf;
        int                                           m_jmaxtf;
        int                                           m_kmaxtf;
        TileIndex                                     m_indexVTTParentCached;
        const uint64_t                                m_texture3DDim;
        const uint64_t                                m_texture3DDim2;
        const bool                                    m_storeTileDataAs3DTextures;
#ifdef OPTIMIZE_USING_PARENT_HALO
        TDH m_tdhParent;
#endif
        explicit Visitor(
            Tree<BuildType>*                in_treeVTT,
            nanovdb::DitherLUT&             in_lut,
            BIFTileAccessor&                in_acc,
            Container*                      in_containerVTT,
            Amino::Vector<uint64_t> const&  in_tdOffset,
            const uint64_t                  in_tdOffsetConstant,
            Amino::Vector<TileIndex> const* in_indexToVTT,
            Amino::Vector<uint64_t> const&      in_tdcOffset,
            Amino::Vector<int> const*           in_tilesToCoalesce,
            const int                           in_rootDepth,
            const int                           in_maxDepth,
            const bool                          in_coalesceTileData,
            Amino::Vector<unsigned char> const* in_tilesToRemove,
            BIF::TileTree3D::
                Channel<BIFTraits, VTTType, BIFTraits::kTileDataWidth> const&
                                                          in_channel,
            Amino::Vector<Amino::Vector<Codec>> const&    in_codecsVec,
            const uint64_t                                in_fpNIndex,
            Amino::Vector<Amino::Vector<uint64_t>> const& in_tdOffsetFpNVec,
            const bool                                    in_isSDF,
            const float                                   in_voxelSize,
            const bool                    in_sortIntoMortonOrder,
            const Amino::Vector<int64_t>* in_permutation,
            const bool                                    in_dilateActiveFogVoxels,
            const float                                   in_activeFogVoxelsThreshold,
            const uint64_t                                in_texture3DDim,
            const bool                                    in_storeTileDataAs3DTextures)
            : m_treeVTT(in_treeVTT),
              m_lut(in_lut),
              m_tdhIn(TDH::kSize),
              m_tdhIn2(nullptr),
              m_dummyCodec(),
              m_acc(in_acc),
              m_containerVTT(in_containerVTT),
              m_tdOffset(in_tdOffset),
              m_tdOffsetConstant(in_tdOffsetConstant),
              m_indexToVTT(in_indexToVTT),
              m_tdcOffset(in_tdcOffset),
              m_tilesToCoalesce(in_tilesToCoalesce),
              m_rootDepth(in_rootDepth),
              m_maxDepth(in_maxDepth),
              m_coalesceTileData(in_coalesceTileData),
              m_tilesToRemove(in_tilesToRemove),
              m_channel(in_channel),
              m_codecsVec(in_codecsVec),
              m_fpNIndex(in_fpNIndex),
              m_tdOffsetFpNVec(in_tdOffsetFpNVec),
              m_isSDF(in_isSDF),
              m_voxelSize(in_voxelSize),
              m_sortIntoMortonOrder(in_sortIntoMortonOrder),
              m_permutation(in_permutation),
              m_dilateActiveFogVoxels(in_dilateActiveFogVoxels),
              m_enableActiveState(m_channel.name() == Bifrost::Geometry::sVoxelFogDensity && m_dilateActiveFogVoxels),
              m_activeFogVoxelsThreshold(in_activeFogVoxelsThreshold),
              m_indexVTTParentCached(kInvalidTileIndex),
              m_texture3DDim(in_texture3DDim / Traits::kExtendedTileDataWidth),
              m_texture3DDim2(m_texture3DDim * m_texture3DDim),
              m_storeTileDataAs3DTextures(in_storeTileDataAs3DTextures)
#ifdef OPTIMIZE_USING_PARENT_HALO
              , m_tdhParent(TDH::kSize)
#endif
        {
            if (m_enableActiveState) {
                m_tdhIn2 = TDH(TDH::kSize);
            }
            m_imin = Traits::kMaxTileCoord;
            m_jmin = Traits::kMaxTileCoord;
            m_kmin = Traits::kMaxTileCoord;
            m_imax = Traits::kMinTileCoord;
            m_jmax = Traits::kMinTileCoord;
            m_kmax = Traits::kMinTileCoord;
            m_imintf = Traits::kMaxTileCoord;
            m_jmintf = Traits::kMaxTileCoord;
            m_kmintf = Traits::kMaxTileCoord;
            m_imaxtf = Traits::kMinTileCoord;
            m_jmaxtf = Traits::kMinTileCoord;
            m_kmaxtf = Traits::kMinTileCoord;
        }

        void beginTile(const BIFTile& in_tile) override {
            const int       depth = in_tile.depth();
            const int       vdw   = in_tile.voxelDataWidth();
            const TileIndex accum = m_containerVTT->m_tileCountAccum[depth];
            // Note that even though the tile has been removed and doesn't have
            // a valid indexVTT we still need to enter here to assign the tile
            // data to the parent tile data.
            const bool tileIsAtMaxDepth = (depth == m_maxDepth);
            const bool tileIsRemoved =
                (m_coalesceTileData
                     ? m_tilesToRemove[in_tile.depth()][in_tile.index()]
                     : false);
            const bool tileIsRemovedLeaf =
                (tileIsRemoved &&
                 ((in_tile.depth()==m_maxDepth) ||
                  m_tilesToCoalesce[in_tile.depth()][in_tile.index()] == 0));
            const TileIndex indexVTT =
                (tileIsRemoved
                     ? std::numeric_limits<TileIndex>::max()
                     : (accum +
                        m_indexToVTT[in_tile.depth()][in_tile.index()]));
            bool isHalo = (tileIsRemoved ? false
                                         : ((m_tdOffset[indexVTT + 1] -
                                             m_tdOffset[indexVTT]) == 1));
            bool computeHalo = isHalo;

            const bool isQuantized = !std::is_same<VoxelT, void>::value &&
                                     std::is_same<VTTType, float>::value;

            bool                tileIsCoalesced = false;
            BIFTile const*      parentTile      = nullptr;
            Bifrost::Math::int3 childIJK{};
            const auto defaultValue = m_channel.defaultValue();

            // If tile is coalesced, its data will be assembled from finer
            // tiles, but we need to intialize stride, shift and offset.
            Bifrost::Math::int3 tdwIJK{
                (isHalo ? Traits::kExtendedTileDataWidth : Traits::kTileDataWidth),
                (isHalo ? Traits::kExtendedTileDataWidth : Traits::kTileDataWidth),
                (isHalo ? Traits::kExtendedTileDataWidth : Traits::kTileDataWidth)};

            auto treeIndex = in_tile.treeIndex();

            // \todo BIFROST-TBD NanoVTT: Support grid's depth
            // to avoid storing data for nothing at the finest levels.

            // The channel and the container don't have the same depth. This
            // concept doesn't exist in nanoVTT, we fill the finest depth with
            // data from coarser depth.
            if (in_tile.depth() > m_channel.depth()) {
                treeIndex  = in_tile.parentTreeIndex();
                parentTile = &m_acc.tile(treeIndex);
                while ( parentTile->depth() > m_channel.depth()) {
                    treeIndex  = parentTile->parentTreeIndex();
                    parentTile = &m_acc.tile(treeIndex);
                }
            }
            parentTile = nullptr;
            m_tdIn                = m_channel.tileData(treeIndex);
            if (!isQuantized && m_coalesceTileData) {
                if (tileIsRemovedLeaf) {
                    // We must find the offset of the parent
                    parentTile = &m_acc.tile(in_tile.parentTreeIndex());

                    while (parentTile->depth() > m_rootDepth &&
                           m_tilesToCoalesce[parentTile->depth() - 1]
                                            [parentTile->parentIndex()] > 0) {
                        parentTile = &m_acc.tile(parentTile->parentTreeIndex());
                    }

                    // Find ijk offset of child into parent.
                    childIJK = {(in_tile.ti() - parentTile->ti()) /
                                    in_tile.voxelDataWidth(),
                                (in_tile.tj() - parentTile->tj()) /
                                    in_tile.voxelDataWidth(),
                                (in_tile.tk() - parentTile->tk()) /
                                    in_tile.voxelDataWidth()};

                    const TileIndex accumParent =
                        m_containerVTT->m_tileCountAccum[parentTile->depth()];
                    const TileIndex parentIndexVTT =
                        accumParent +
                        m_indexToVTT[parentTile->depth()][parentTile->index()];

                    const bool parentIsHalo =
                        ((m_tdOffset[parentIndexVTT + 1] -
                          m_tdOffset[parentIndexVTT]) == 1);

                    if (parentIsHalo != isHalo) {
                        // (Coalesced) Parent halo status overrides child halo
                        // status
                        isHalo = parentIsHalo;
                        tdwIJK = {(isHalo ? Traits::kExtendedTileDataWidth
                                          : Traits::kTileDataWidth),
                                  (isHalo ? Traits::kExtendedTileDataWidth
                                          : Traits::kTileDataWidth),
                                  (isHalo ? Traits::kExtendedTileDataWidth
                                          : Traits::kTileDataWidth)};
                    }

                    computeHalo = isHalo;
                    if (isHalo) {
                        int numAdjustedHaloEdges = 0;
                        if (in_tile.ti() + in_tile.tileVoxelWidth() <
                            parentTile->ti() + parentTile->tileVoxelWidth()) {
                            ++numAdjustedHaloEdges;
                            tdwIJK.x = Traits::kTileDataWidth;
                        }
                        if (in_tile.tj() + in_tile.tileVoxelWidth() <
                            parentTile->tj() + parentTile->tileVoxelWidth()) {
                            ++numAdjustedHaloEdges;
                            tdwIJK.y = Traits::kTileDataWidth;
                        }
                        if (in_tile.tk() + in_tile.tileVoxelWidth() <
                            parentTile->tk() + parentTile->tileVoxelWidth()) {
                            ++numAdjustedHaloEdges;
                            tdwIJK.z = Traits::kTileDataWidth;
                        }
                        // If we're not at the edge of the parent tile there's
                        // no need to compute the halo.
                        if (numAdjustedHaloEdges == 3) computeHalo = false;
                    }
                } // if (tileIsRemovedLeaf) ...
                else {
                    tileIsCoalesced =
                        !(in_tile.depth()==m_maxDepth) &&
                        m_tilesToCoalesce[in_tile.depth()][in_tile.index()] >
                            0 &&
                        (in_tile.depth() == m_rootDepth ||
                         m_tilesToCoalesce[in_tile.depth() - 1]
                                          [in_tile.parentIndex()] == 0);

                    if (tileIsCoalesced) {
                        const int tileDataShift =
                            m_tilesToCoalesce[in_tile.depth()][in_tile.index()];
                        const int tileDataStride =
                            static_cast<int>(
                                Traits::kTileDataWidth *
                                std::pow(Traits::kTileWidth, tileDataShift)) +
                            static_cast<int>(isHalo ? 1 : 0);

                        initAddDataToTile<BuildType>(
                            indexVTT,
                            m_tdOffsetConstant +
                                m_tdcOffset[indexVTT] * valueTypeSize,
                            tileDataStride, tileDataShift, *m_treeVTT);

                        return;
                    } // if (tileIsCoalesced)
                }     // if (tileIsRemoved) ... else ...
            }         // if (!isQuantized && m_coalesceTileData)
            if (computeHalo) {
                const auto haloMode = BIF::TileTree3D::HaloMode::kInterpolateLinear;
#ifdef OPTIMIZE_USING_PARENT_HALO
                // \todo BIFROST-TBD - Currently disabled for quantization as this
                // would require extra work to unpack the data.
                bool parentIsHalo = (depth > rootDepth && !isQuantized);
                TileIndex indexVTTParent = kInvalidTileIndex;
                if (parentIsHalo) {
                    const TileIndex accumParent =
                        m_containerVTT->m_tileCountAccum[depth - 1];

                    indexVTTParent =
                        accumParent + m_indexToVTT[in_tile.depth() - 1]
                                                  [in_tile.parentIndex()];
                    parentIsHalo = ((m_tdOffset[indexVTTParent + 1] -
                                     m_tdOffset[indexVTTParent]) == 1);
                }

                if (parentIsHalo) {
                    // Here we construct m_tdhParent from the NanoVTT and re-use.

                    if (parentTile == nullptr) {
                        parentTile = &m_acc.tile(in_tile.parentTreeIndex());
                    }

                    // Fill in the parent tile data
                    // Here we check for a cached parent to avoid copying the parent tile data again.
                    if (m_indexVTTParentCached != indexVTTParent) {
                        m_indexVTTParentCached = indexVTTParent;
                        if (m_storeTileDataAs3DTextures) {
                            const uint64_t tvDim = m_texture3DDim * Traits::kExtendedTileDataWidth;
                            const uint64_t tvDim2 = tvDim * tvDim;

                            // Compute the 3D texture coordinate from the index
                            // Here we add 1 to account for a tile of default values.
                            const uint64_t textureIndex = indexVTTParent + 1;
                            const uint64_t x = (textureIndex % m_texture3DDim);
                            const uint64_t z = (textureIndex / m_texture3DDim2);
                            const uint64_t y = ((textureIndex - m_texture3DDim2 * z) / m_texture3DDim);
                            using ValueType = typename Tree<BuildType>::ValueType;
                            // Base pointer to texture
                            ValueType* io_texture = m_treeVTT->tileDataFromOffset(m_tdOffsetConstant);
                            int ii = 0;
                            for (uint64_t k = 0; k < Traits::kExtendedTileDataWidth; ++k)
                                for (uint64_t j = 0; j < Traits::kExtendedTileDataWidth; ++j)
                                    for (uint64_t i = 0; i < Traits::kExtendedTileDataWidth;
                                        ++i, ++ii)
                            {
                                const int64_t textureIndex2 =
                                    (z * Traits::kExtendedTileDataWidth + k) *
                                    tvDim2 +
                                    (y * Traits::kExtendedTileDataWidth + j) *
                                    tvDim +
                                    (x * Traits::kExtendedTileDataWidth + i);

                                m_tdhParent(ii) = *reinterpret_cast<VTTType const*>(&io_texture[textureIndex2]);
                            }
                        }
                        else {
                            // \todo BIFROST-TBD - Also support non-3d-texture storage format
                        }
                    } // if (m_indexVTTParentCached != indexVTTParent)
                    // Computes a halo tile data of size (N+1)^3 for a tile of size (N)^3
                    // and interpolates from the known parent tile data. This saves computation time.
                    BIF::TileTree3D::setHaloTile<BIFTraits, VTTType, 0, 0, 1,
                                                 0>(in_tile, parentTile,
                                                    m_channel, m_tdhParent,
                                                    haloMode, m_acc, m_tdhIn);
                }
                else
#endif
                {
                    BIF::TileTree3D::setHaloTile<BIFTraits, VTTType, 0, 0, 0,
                        1>(
                            in_tile.treeIndex(), m_channel, haloMode,
                            BIF::TileTree3D::HaloType::kAll, m_acc, m_tdhIn);
                }

                if (kClampSDFVoxelValues)
                    clampSDFVoxelValues(defaultValue, m_tdhIn, kClampSDFVoxelValuesThreshold * m_voxelSize);

                if (m_enableActiveState) {
                    // Here we make sure to dilate the active set of fog
                    // voxels slightly within a tile such that we can still
                    // use the voxel-based HDDA without artifacts. Note:
                    // this needs to be done at all levels, not just the
                    // finest level, to make sure that ray marching also
                    // works when we use coarser levels.
                    setActiveState(defaultValue, m_tdhIn, m_tdhIn2,
                                   m_activeFogVoxelsThreshold);
                } // if (m_enableActiveState)

            } // if (computeHalo)
            // \todo BIFROST-TBD - Also enable in the case where computeHalo is false.
            const auto& tdhIn = ((m_enableActiveState && computeHalo) ? m_tdhIn2 : m_tdhIn);
            if (!isQuantized) {
                if (m_coalesceTileData) {
                    int64_t offset;
                    if (tileIsRemoved) {
                        if (tileIsRemovedLeaf) {
                            const TileIndex accumParent =
                                m_containerVTT
                                    ->m_tileCountAccum[parentTile->depth()];
                            const TileIndex parentIndexVTT =
                                accumParent + m_indexToVTT[parentTile->depth()]
                                                          [parentTile->index()];

                            const int tileDataShift =
                                m_tilesToCoalesce[parentTile->depth()]
                                                 [parentTile->index()];
                            const int tileDataStride =
                                static_cast<int>(Traits::kTileDataWidth *
                                                 std::pow(Traits::kTileWidth,
                                                          tileDataShift)) +
                                static_cast<int>(isHalo ? 1 : 0);

                            if (computeHalo) {
                                addDataToTile<BuildType>(
                                    &tdhIn(0), m_lut, parentIndexVTT,
                                    m_dummyCodec, Traits::kExtendedTileDataWidth,
                                    tileDataStride,
                                    {childIJK.x, childIJK.y, childIJK.z},
                                    {tdwIJK.x, tdwIJK.y, tdwIJK.z}, *m_treeVTT);
                            } else {
                                addDataToTile<BuildType>(
                                    &m_tdIn(0), m_lut, parentIndexVTT,
                                    m_dummyCodec, Traits::kTileDataWidth,
                                    tileDataStride,
                                    {childIJK.x, childIJK.y, childIJK.z},
                                    {tdwIJK.x, tdwIJK.y, tdwIJK.z}, *m_treeVTT);
                            }
                        } // if (tileIsRemovedLeaf)
                    }     // if (tileIsRemoved)
                    else {
                        offset = m_tdcOffset[indexVTT] * valueTypeSize;

                        if (isHalo) {
                            addDataToTile<BuildType>(
                                &tdhIn(0), m_lut, indexVTT, m_dummyCodec,
                                offset + m_tdOffsetConstant, *m_treeVTT,
                                Traits::kExtendedTileDataSize);
                        } else {

                            addDataToTile<BuildType>(
                                &m_tdIn(0), m_lut, indexVTT, m_dummyCodec,
                                offset + m_tdOffsetConstant, *m_treeVTT,
                                Traits::kTileDataSize);
                        }
                    } //  if (tileIsRemoved) ... else ...
                }     // if (m_coalesceTileData)
                else {
                    if (m_storeTileDataAs3DTextures) {
                        assert(isHalo);
                        const uint64_t tvDim = m_texture3DDim * Traits::kExtendedTileDataWidth;
                        const uint64_t tvDim2 = tvDim * tvDim;

                        // Compute the 3D texture coordinate from the index
                        // Here we add 1 to account for a tile of default values.
                        const uint64_t textureIndex = indexVTT + 1;
                        const uint64_t x = (textureIndex % m_texture3DDim);
                        const uint64_t z = (textureIndex / m_texture3DDim2);
                        const uint64_t y = ((textureIndex - m_texture3DDim2 * z) / m_texture3DDim);
                        const uint64_t offset =
                            ((x * Traits::kExtendedTileDataWidth) |
                             ((((z * Traits::kExtendedTileDataWidth)
                                << Traits::k3DTextureShift) |
                               (y * Traits::kExtendedTileDataWidth))
                              << Traits::k3DTextureShift));

                        m_treeVTT->setTileDataOffset(indexVTT, offset);
                        m_treeVTT->setTileDataStride(indexVTT, static_cast<uint8_t>(Traits::kExtendedTileDataWidth));
                        m_treeVTT->setTileDataShift(indexVTT, 0);
                        using ValueType = typename Tree<BuildType>::ValueType;
                        // Base pointer to texture
                        ValueType* io_texture = m_treeVTT->tileDataFromOffset(m_tdOffsetConstant);
                        int ii = 0;
                        for (uint64_t k = 0; k < Traits::kExtendedTileDataWidth; ++k)
                            for (uint64_t j = 0; j < Traits::kExtendedTileDataWidth; ++j)
                                for (uint64_t i = 0; i < Traits::kExtendedTileDataWidth;
                                     ++i, ++ii)
                        {
                                    const int64_t textureIndex2 =
                                        (z * Traits::kExtendedTileDataWidth + k) *
                                            tvDim2 +
                                        (y * Traits::kExtendedTileDataWidth + j) *
                                            tvDim +
                                        (x * Traits::kExtendedTileDataWidth + i);
                            io_texture[textureIndex2] =
                                *reinterpret_cast<ValueType const*>(&tdhIn(ii));
                        }

                        // If this is the first proper tile, we also initialize
                        // the default tile.
                        if (indexVTT == 0) {
                            for (uint64_t k = 0; k < Traits::kExtendedTileDataWidth;
                                 ++k)
                                for (uint64_t j = 0;
                                     j < Traits::kExtendedTileDataWidth; ++j)
                                    for (uint64_t i = 0;
                                         i < Traits::kExtendedTileDataWidth; ++i) {
                                        const int64_t textureIndex2 =
                                            k * tvDim2 + j * tvDim + i;
                                        io_texture[textureIndex2] =
                                            *reinterpret_cast<ValueType const*>(
                                                &defaultValue);
                                    }
                        }

                    } // if (m_storeTileDataAs3DTextures)
                    else {
                        assert(computeHalo == isHalo);
                        const int64_t offset =
                            (indexVTT == 0 ? 0
                                : (m_tdOffset[indexVTT] *
                                    Traits::kExtendedTileDataSize +
                                    (indexVTT - m_tdOffset[indexVTT]) *
                                    Traits::kTileDataSize) *
                                valueTypeSize);
                        if (isHalo) {
                            addDataToTile<BuildType>(
                                &tdhIn(0), m_lut, indexVTT, m_dummyCodec,
                                offset + m_tdOffsetConstant, *m_treeVTT,
                                Traits::kExtendedTileDataSize);
                        }
                        else {
                            addDataToTile<BuildType>(
                                &m_tdIn(0), m_lut, indexVTT, m_dummyCodec,
                                offset + m_tdOffsetConstant, *m_treeVTT);
                        }
                    }
                } // if (m_coalesceTileData) ... else ...
            } else if (std::is_same<BuildType, nanovdb::FpN>::value) {
                if (isHalo) {
                    addDataToTile<BuildType>(
                        &tdhIn(0), m_lut, indexVTT,
                        m_codecsVec[m_fpNIndex][indexVTT],
                        m_tdOffsetFpNVec[m_fpNIndex][indexVTT], *m_treeVTT,
                        Traits::kExtendedTileDataSize);
                } else {
                    addDataToTile<BuildType>(
                        &m_tdIn(0), m_lut, indexVTT,
                        m_codecsVec[m_fpNIndex][indexVTT],
                        m_tdOffsetFpNVec[m_fpNIndex][indexVTT], *m_treeVTT);
                }
            } else {
                // Fp4, Fp8, Fp16
                const int64_t offset =
                    (indexVTT == 0 ? 0
                                   : (m_tdOffset[indexVTT] *
                                          Tree<BuildType>::kExtendedTileDataBytes +
                                      (indexVTT - m_tdOffset[indexVTT]) *
                                          Tree<BuildType>::kTileDataBytes));
                if (isHalo) {
                    addDataToTile(&tdhIn(0), m_lut, indexVTT, m_dummyCodec,
                                  offset + m_tdOffsetConstant, *m_treeVTT,
                                  Traits::kExtendedTileDataSize);
                } else {
                    addDataToTile(&m_tdIn(0), m_lut, indexVTT, m_dummyCodec,
                                  offset + m_tdOffsetConstant, *m_treeVTT);
                }
            }

            // Build world bbox.
            int           iminl           = Traits::kTileDataWidth;
            int           jminl           = Traits::kTileDataWidth;
            int           kminl           = Traits::kTileDataWidth;
            int           imaxl           = -1;
            int           jmaxl           = -1;
            int           kmaxl           = -1;
            // Tight fitting bbox
            int           iminltf = Traits::kTileDataWidth;
            int           jminltf = Traits::kTileDataWidth;
            int           kminltf = Traits::kTileDataWidth;
            int           imaxltf = -1;
            int           jmaxltf = -1;
            int           kmaxltf = -1;
            bool          foundInsideBBox = false;
            bool          foundInsideTightFitBBox = false;
            constexpr int tdwTwFactor =
                Traits::kTileDataWidth / Traits::kTileWidth;
            // Note: one can also use 'm_voxelSize * vdw * 1.73205080757f' to capture
            //       parts that are not negative, but close. However, this results in a
            //       larger bbox and hence more expensive rendering.
            const float sdfThreshold = 0.f;

            for (int k = 0; k < Traits::kTileWidth; ++k) {
                for (int j = 0; j < Traits::kTileWidth; ++j) {
                    for (int i = 0; i < Traits::kTileWidth; ++i) {
                        bool foundInsideTightFitBBoxLocal = false;
                        if (tileIsAtMaxDepth || in_tile.isLeaf(i, j, k)) {
                            for (int k2 = k * tdwTwFactor; k2 < (k + 1) * tdwTwFactor; ++k2) {
                                for (int j2 = j * tdwTwFactor; j2 < (j + 1) * tdwTwFactor; ++j2) {
                                    for (int i2 = i * tdwTwFactor; i2 < (i + 1) * tdwTwFactor; ++i2) {
                                        if (isSignificant(
                                            m_isSDF, sdfThreshold, defaultValue,
                                            (isHalo ? tdhIn(i2, j2, k2) : m_tdIn(i2, j2, k2)),
                                            (m_enableActiveState
                                                ? &foundInsideTightFitBBoxLocal
                                                : nullptr),
                                            m_activeFogVoxelsThreshold)) {
                                            iminl = std::min(iminl, i2);
                                            jminl = std::min(jminl, j2);
                                            kminl = std::min(kminl, k2);
                                            imaxl = std::max(imaxl, i2);
                                            jmaxl = std::max(jmaxl, j2);
                                            kmaxl = std::max(kmaxl, k2);
                                            foundInsideBBox = true;
                                            if (foundInsideTightFitBBoxLocal) {
                                                iminltf = std::min(iminltf, i2);
                                                jminltf = std::min(jminltf, j2);
                                                kminltf = std::min(kminltf, k2);
                                                imaxltf = std::max(imaxltf, i2);
                                                jmaxltf = std::max(jmaxltf, j2);
                                                kmaxltf = std::max(kmaxltf, k2);
                                                foundInsideTightFitBBox = foundInsideTightFitBBoxLocal;
                                            }
                                        }
                                    }
                                }
                            } // for (int kk = 0; kk < tdwTwFactor; ++kk)
                        } // if (tileIsAtMaxDepth || in_tile.isLeaf(i, j, k)
                    }
                }
            } // for (int k = 0; k < Traits::kTileWidth; ++k)

            if (foundInsideBBox) {
                // We pad the bbox with two voxels at the current level (BIFROST-4711).
                constexpr int bboxPadding = 2;
                iminl -= bboxPadding;
                jminl -= bboxPadding;
                kminl -= bboxPadding;
                imaxl += bboxPadding;
                jmaxl += bboxPadding;
                kmaxl += bboxPadding;

                iminl *= vdw;
                imaxl *= vdw;
                jminl *= vdw;
                jmaxl *= vdw;
                kminl *= vdw;
                kmaxl *= vdw;

                iminl += in_tile.ti();
                imaxl += in_tile.ti();
                jminl += in_tile.tj();
                jmaxl += in_tile.tj();
                kminl += in_tile.tk();
                kmaxl += in_tile.tk();

                m_imin = std::min(iminl, m_imin);
                m_jmin = std::min(jminl, m_jmin);
                m_kmin = std::min(kminl, m_kmin);
                m_imax = std::max(imaxl, m_imax);
                m_jmax = std::max(jmaxl, m_jmax);
                m_kmax = std::max(kmaxl, m_kmax);
            }
            if (foundInsideTightFitBBox) {
                // Here we pad slightly to avoid cutting off the data when displayed
                constexpr int bboxPadding = 2;
                iminltf -= bboxPadding;
                jminltf -= bboxPadding;
                kminltf -= bboxPadding;
                imaxltf += bboxPadding;
                jmaxltf += bboxPadding;
                kmaxltf += bboxPadding;

                iminltf *= vdw;
                imaxltf *= vdw;
                jminltf *= vdw;
                jmaxltf *= vdw;
                kminltf *= vdw;
                kmaxltf *= vdw;

                iminltf += in_tile.ti();
                imaxltf += in_tile.ti();
                jminltf += in_tile.tj();
                jmaxltf += in_tile.tj();
                kminltf += in_tile.tk();
                kmaxltf += in_tile.tk();

                m_imintf = std::min(iminltf, m_imintf);
                m_jmintf = std::min(jminltf, m_jmintf);
                m_kmintf = std::min(kminltf, m_kmintf);
                m_imaxtf = std::max(imaxltf, m_imaxtf);
                m_jmaxtf = std::max(jmaxltf, m_jmaxtf);
                m_kmaxtf = std::max(kmaxltf, m_kmaxtf);
            }
        }
        bool visitInteriorVoxel(const BIFTile&,
                                const Bifrost::Math::int3&) override {
            return true;
        }
        void join(const BIFContainer::VisitorType& in) override {
            const auto& in2 = *static_cast<Visitor const*>(&in);
            m_imin          = std::min(in2.m_imin, m_imin);
            m_jmin          = std::min(in2.m_jmin, m_jmin);
            m_kmin          = std::min(in2.m_kmin, m_kmin);
            m_imax          = std::max(in2.m_imax, m_imax);
            m_jmax          = std::max(in2.m_jmax, m_jmax);
            m_kmax          = std::max(in2.m_kmax, m_kmax);
            m_imintf = std::min(in2.m_imintf, m_imintf);
            m_jmintf = std::min(in2.m_jmintf, m_jmintf);
            m_kmintf = std::min(in2.m_kmintf, m_kmintf);
            m_imaxtf = std::max(in2.m_imaxtf, m_imaxtf);
            m_jmaxtf = std::max(in2.m_jmaxtf, m_jmaxtf);
            m_kmaxtf = std::max(in2.m_kmaxtf, m_kmaxtf);
        }
        BIFContainer::VisitorType* copy() override {
            // Note: it's not necessary to clear the reduction variables
            //       because we're doing min, max operations only.
            return new Visitor(*this);
        }
    };

    constexpr bool verboseTimings = false;
    constexpr bool outputTraits28StorageInfo = false;
    constexpr bool outputTileDataCoarseningStorageInfo = false;
    constexpr bool outputMortonOrderStorageInfo = false;
    constexpr int traversalStructureElementCount = 4;
    constexpr int traversalStructureElementSize = traversalStructureElementCount * sizeof(int);
    constexpr int neighborStructureElementSize = sizeof(TreeIndex) * 27;

    template<typename TileT, typename AccT>
    bool checkGraded(const TileT& in_tile, AccT& io_acc, int& io_numHaloPoints)
    {
        constexpr int haloSumTable[] = { 0, Traits::kTileWidth2, Traits::kTileWidth, 1 };
        io_numHaloPoints = 0;
        const int depth = in_tile.depth();
        bool graded = true;
        for (int i = 1; i <= 2 && graded; ++i) for (int j = 1; j <= 2 && graded; ++j) for (int k = 1; k <= 2 && graded; ++k) {
            const BIF::TileTree3D::NeighborDirection nbDir = static_cast<BIF::TileTree3D::NeighborDirection>(i * 9 + j * 3 + k);
            if (nbDir != BIF::TileTree3D::NeighborDirection::kCenter) {
                const int sum = i + j + k - 3;
                auto& tnb = io_acc.tileNeighbor(in_tile, nbDir);
                if (tnb.valid()) {
                    if (tnb.depth() < depth - 1) {
                        graded = false;
                        io_numHaloPoints += haloSumTable[sum];
                    }
                    else if (tnb.depth() < depth) {
                        io_numHaloPoints += haloSumTable[sum];
                    }
                }
                // If tnb is not valid we will just return the default value which is not expensive.
            }
        }
        return graded;
    }

    template <typename BufferT, typename VoxelT, typename OracleT, bool kEnableCoalescedTilesSupportT>
    GridHandle<BufferT> vttToNanoVTT(const BIF::Object& in_volume,            // NOLINT
                                     const Amino::String& in_properties,
                                     const nanovdb::StatsMode /*in_sMode*/,
                                     const nanovdb::ChecksumMode in_cMode,
                                     const int                   in_verbose,
                                     const float                 in_tolerance,
                                     const bool                  in_ditherOn,
                                     const float    in_extendedBudgetFraction,
                                     const int      in_minAdaptiveLevels,
                                     const int      in_discardFinestLevels,
                                     const bool     in_sortIntoMortonOrder,
                                     const bool     in_coalesceTileData,
                                     const bool     in_padBufferForVolumetricTexture,
                                     const int      in_textureElementSize,
                                     const bool     in_storeTileDataAs3DTextures,
                                     const bool     in_dilateActiveFogVoxels,
                                     const float    in_activeFogVoxelsThreshold,
                                     const bool     in_setCoarserLevels,
                                     const bool     in_enableNanoVTTTraversal,
                                     BufferT*       out_traversalStructure,
                                     BufferT*       out_neighborStructure,
                                     Bifrost::Math::bboxf* out_bbox,
                                     const BufferT& in_allocator) {
        const bool verbose = (in_verbose > 0);
        tbb::tick_count startTime2, endTime2;

#if 0
        std::cout << "Start profiling" << std::endl;
        std::this_thread::sleep_for(std::chrono::milliseconds(25000));
        std::cout << "Profile started" << std::endl;
#endif

        const int kMaxDepth = Traits::kMaxDepth - std::max(std::min(in_discardFinestLevels, Traits::kMaxDepth), 0);
        // NANOVTT_SUPPORT_COALESCED_TILES off, and coalesceTileData on:
        // - Internal tiles with all children refined are by-passed by the parent-child pointers (but not removed)
        //   and their tile data size are set to zero.
        // - All non-bypassed tile data is still the same size (4^3).
        // - Parent tiles to coalesced tiles will have more child pointers and therefore be larger than 4^3.
        // NANOVTT_SUPPORT_COALESCED_TILES on, and coalesceTileData on:
        // - By-passed tiles and leaves are actually removed.
        // - Tile data is also coalesced into larger tile datas (4^3, 8^3 etc).
        // \todo BIFROST-TBD Support for quantized values with coalesced tiles.
        const bool coalesceTileData = (std::is_same<VoxelT, void>::value) && in_coalesceTileData;
        if (coalesceTileData != in_coalesceTileData) {
            std::cerr << "VTTToNanoVTT: coalesceTileData disabled. Requirements: currently not supported with quantization." << std::endl;
        }
        const bool storeTileDataAs3DTextures = (std::is_same<VoxelT, void>::value) && in_storeTileDataAs3DTextures && !coalesceTileData && (in_extendedBudgetFraction >= 1.f);
        if (storeTileDataAs3DTextures != in_storeTileDataAs3DTextures) {
            std::cerr << "VTTToNanoVTT: storeTileDataAs3DTextures disabled. Requirements: currently not supported with quantization nor tile data coalescing." << std::endl;
        }
        const bool enableNanoVTTTraversal = (in_enableNanoVTTTraversal && out_traversalStructure != nullptr && !coalesceTileData);
        if (enableNanoVTTTraversal != in_enableNanoVTTTraversal) {
            std::cerr << "VTTToNanoVTT: enableNanoVTTTraversal disabled. Requirements: out_traversalStructure != nullptr and coalesceTileData == false." << std::endl;
        }
        const bool enableNanoVTTNeighbors = (out_neighborStructure != nullptr && !coalesceTileData);
        if (!enableNanoVTTNeighbors && out_neighborStructure != nullptr) {
            std::cerr << "VTTToNanoVTT: enableNanoVTTNeighbors disabled. Requirements: out_neighborStructure != nullptr and coalesceTileData == false." << std::endl;
        }
        const bool setCoarserLevels = (in_setCoarserLevels && !std::is_same<VoxelT, nanovdb::FpN>::value);
        if (setCoarserLevels != in_setCoarserLevels) {
            std::cerr << "VTTToNanoVTT: setCoarserLevels disabled. Requirements: VoxelT != nanovdb::FpN." << std::endl;
        }

        if (coalesceTileData) {
            if (verbose)
                std::cout << "coalesceTileData=ON, NANOVTT_SUPPORT_COALESCED_TILES=";
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
            assert(kEnableCoalescedTilesSupportT);
            if (verbose)
                std::cout << "ON";
#else
            assert(!kEnableCoalescedTilesSupportT);
            if (verbose)
                std::cout << "OFF";
#endif
            if (verbose)
                std::cout << ", kEnableCoalescedTilesSupportT = " << (kEnableCoalescedTilesSupportT ? "ON" : "OFF") << std::endl;
        }


        tbb::tick_count startTime, endTime;
        startTime = tbb::tick_count::now();
        startTime2 = tbb::tick_count::now();

        GridHandle<BufferT> handle;

        if (auto container = in_volume.component<BIFContainer>(
            Bifrost::Geometry::sVoxelTileTree))
            if (auto voxelTopology = in_volume.component<BIFVoxelTopology>(
                Bifrost::Geometry::sVoxelComp)) {
                // First pass:
                // Here we compute the storage requirements and iterate over all
                // component names to see if it's a channel of the given type that
                // should be stored in the buffer.

                // CONTAINER.BEGIN (pass 1)

                Amino::Vector<int64_t> permutation[BIFTraits::kMaxDepth + 1];
                if (in_sortIntoMortonOrder) {
                    // The permutation gives the mapping FROM existing linear index 'i' at a given depth TO the index where
                    // the corresponding value will end up 'iPermuted':
                    // iPermuted = permutation[depth][i];
                    Amino::Vector<BIF::TileTree3D::MortonCode> mortonOrder[BIFTraits::kMaxDepth + 1];
                    BIF::TileTree3D::initializePermutation(*container, permutation);
                    BIF::TileTree3D::computeMortonOrder(*container, mortonOrder);
                    BIF::TileTree3D::sortIntoOrder<BIFTraits>(mortonOrder, permutation);
                }

                // Determine root depth and hash table
                int rootDepth = 0;
                nanovdb::Coord hashTableMin;
                bool done = false;
                for (int depth = Traits::kMaxDepth - std::max(std::min(std::max(in_minAdaptiveLevels, in_discardFinestLevels), Traits::kMaxDepth), 0); depth >= 0 && !done; --depth) {
                    Bifrost::Math::bboxi bbox;
                    if (container->tileCount(depth) > 0) {
                        container->tileSpaceBounds(depth, bbox);
                        auto tileSpaceSize = bbox.size();
                        int tvw = container->tileVoxelWidth(depth);
                        int maxTileSpaceSize = std::max(tileSpaceSize.x, std::max(tileSpaceSize.y, tileSpaceSize.z));
                        int maxTileSpaceEntries = maxTileSpaceSize / tvw;
                        if (maxTileSpaceEntries <= Container::kTableWidth) {
                            // In this case the acceleration structure can be introduced at the current depth.
                            done = true;

                            // Here we make sure that we do not discard adaptive
                            // data for fog and level sets by retaining
                            // significant leaf data.
                            tbb::atomic<int> rootDepthReduction = depth;
                            {
                                using BIFChannel = BIF::TileTree3D::Channel<
                                    BIFTraits, float,
                                    BIFTraits::kTileDataWidth>;
                                auto fogChannel =
                                    in_volume.component<BIFChannel>(
                                        Bifrost::Geometry::sVoxelFogDensity);
                                auto sdfChannel = in_volume.component<
                                    BIFChannel>(
                                    Bifrost::Geometry::sVoxelSignedDistance);
                                const bool  hasFog = (fogChannel != nullptr);
                                const bool  hasSDF = (sdfChannel != nullptr);
                                const float defaultFog =
                                    (hasFog ? fogChannel->defaultValue() : 0.f);
                                const float defaultSDF =
                                    (hasSDF ? sdfChannel->defaultValue() : 0.f);

                                if (hasFog || hasSDF) {
                                    container->traverse(
                                        [&](const auto& in_tile) {
                                            constexpr int tdwTwFactor =
                                                Traits::kTileDataWidth /
                                                Traits::kTileWidth;
                                            // No need to consider this tile if
                                            // we already know this depth-level
                                            // is significant.
                                            if (rootDepthReduction <=
                                                in_tile.depth())
                                                return;

                                            // Note: one can also use
                                            // 'm_voxelSize * vdw
                                            // * 1.73205080757f' to capture
                                            //       parts that are not
                                            //       negative, but close.
                                            //       However, this results in a
                                            //       larger bbox and hence more
                                            //       expensive rendering.
                                            const float sdfThreshold = 0.f;

                                            using TD =
                                                BIF::TileTree3D::ConstTileData<
                                                    float,
                                                    Traits::kTileDataWidth>;

                                            const TD tdFog =
                                                (hasFog
                                                     ? fogChannel->tileData(
                                                           in_tile.treeIndex())
                                                     : nullptr);
                                            const TD tdSDF =
                                                (hasSDF
                                                     ? sdfChannel->tileData(
                                                           in_tile.treeIndex())
                                                     : nullptr);

                                            for (int k = 0;
                                                 k < Traits::kTileDataWidth;
                                                 ++k)
                                                for (int j = 0;
                                                     j < Traits::kTileDataWidth;
                                                     ++j)
                                                    for (int i = 0;
                                                         i <
                                                         Traits::kTileDataWidth;
                                                         ++i) {
                                                        // No need for separate check for maxDepth
                                                        // here because we're only traversing coarser
                                                        // depths.
                                                        if (in_tile.isLeaf(
                                                                i / tdwTwFactor,
                                                                j / tdwTwFactor,
                                                                k / tdwTwFactor) &&
                                                            ((hasFog &&
                                                              isSignificant(
                                                                  /*isSDF*/
                                                                  false,
                                                                  /*sdfThreshold*/
                                                                  0.f,
                                                                  defaultFog,
                                                                  tdFog(i, j,
                                                                        k), nullptr, 0.f)) ||
                                                             (hasSDF &&
                                                              isSignificant(
                                                                  /*isSDF*/
                                                                  true,
                                                                  sdfThreshold,
                                                                  defaultSDF,
                                                                  tdSDF(i, j,
                                                                        k), nullptr, 0.f)))) {
                                                            rootDepthReduction =
                                                                in_tile.depth();
                                                            return;
                                                        }
                                                    }
                                        },
                                        BIF::TileTree3D::TraversalMode::
                                            kParallelBreadthFirstBottomUp,
                                        BIF::TileTree3D::kInvalidTreeIndex, 0,
                                        depth - 1);
                                } // if (hasFog || hasSDF)
                            }
                            rootDepth = rootDepthReduction;

                            container->tileSpaceBounds(rootDepth, bbox);
                            const auto llc = bbox.min();
                            hashTableMin = *reinterpret_cast<const nanovdb::Coord*>(&llc);
                            tileSpaceSize = bbox.size();
                            tvw = container->tileVoxelWidth(depth);
                            maxTileSpaceSize = std::max(tileSpaceSize.x, std::max(tileSpaceSize.y, tileSpaceSize.z));
                            maxTileSpaceEntries = maxTileSpaceSize / tvw;

                            if (verbose)
                            {
                                std::cout << "Building perfect hash table at depth " << depth << std::endl;
                                std::cout << " maxTileSpaceSize = " << maxTileSpaceSize << std::endl;
                                std::cout << " maxTileSpaceEntries = " << maxTileSpaceEntries << std::endl;
                                std::cout << " llc = " << llc << std::endl;
                                std::cout << " tvw = " << tvw << std::endl;
                            }
                        }
                    }
                } // for (int depth = Traits::kMaxDepth - in_minAdaptiveLevels; depth >= 1 && !done; --depth)
                if (verbose)
                {
                    std::cout << "rootDepth = " << rootDepth << std::endl;
                    std::cout << "number of roots = "
                        << container->tileCount(rootDepth) << std::endl;
                    std::cout << "maxRoots = " << Container::kTableSize << std::endl;
                }

                // Now check if there are any roots at finer depths. That is currently not supported by NanoVTT.
                {
                    auto iter = container->cbeginRoots();
                    auto iend = container->cendRoots();
                    while (iter != iend) {
                        const auto treeIndex = iter->second;
                        if (treeIndex.m_depth > rootDepth) {
                            std::cerr << "VTTToNanoVTT: roots at depths finer than rootDepth (in this case " << rootDepth << " not supported)" << std::endl;
                            return handle;
                        }
                        ++iter;
                    }
                }

                // indexToVTT is a mapping from in_tile.index() to indexVTT
                Amino::Vector<TileIndex> indexToVTT[Traits::kMaxDepth + 1];
                for (int d = rootDepth; d <= kMaxDepth; ++d) {
                    indexToVTT[d].resize(container->tileCount(d), 0);
                }
                // First compute the mapping from VTT to NanoVTT indices (including the Morton permutation).
                container->traverse(
                    [&](const auto& in_tile) {
                        const int       depth = in_tile.depth();
                        const int64_t   index = in_tile.index();
                        const TileIndex indexVTT = static_cast<TileIndex>(
                            in_sortIntoMortonOrder ? permutation[depth][index]
                            : index);
                        indexToVTT[depth][index] = indexVTT;
                    },
                    BIF::TileTree3D::TraversalMode::kParallel,
                        BIF::TileTree3D::kInvalidTreeIndex, rootDepth,
                        kMaxDepth);

                // tilesToCoalesce is non-zero if tiles (and tiledata) below it can
                // be coalesced into a larger tile. The value of tilesToCoalesce denotes
                // the number of depth-levels that can be coalesced. If a value of 1, then
                // one depth level of 4^3 tile data can be coalesced into an 8^3 tile data, if 2
                // then two depth levels of 4^3 tile data can be coalesced int a 16^3 tile data etc.
                Amino::Vector<int> tilesToCoalesce[Traits::kMaxDepth];
                // Note: it's not thread safe to write to distinct elements in a bool vector,
                //       so we must use chars here.
                Amino::Vector<unsigned char> tilesToRemove[Traits::kMaxDepth + 1];
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                // tilesMaxChildDepth holds for depth<Traits::kMaxDepth the
                // max depth of any of the children.
                Amino::Vector<int> tilesMaxChildDepth[Traits::kMaxDepth];
#endif
                int64_t tileCountAccum[Traits::kMaxDepth + 2];
                if (coalesceTileData)
                {
                    auto acc = container->tileAccessor();
                    // Set default values to zero
                    for (int d = rootDepth; d <= kMaxDepth; ++d) {
                        if (d < kMaxDepth) {
                            tilesToCoalesce[d].resize(container->tileCount(d), 0);
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                            tilesMaxChildDepth[d].resize(container->tileCount(d), 0);
#endif
                        }
                        tilesToRemove[d].resize(container->tileCount(d), 0);
                    }

                    // Bottom up traversal
                    container->traverse(
                        [&, acc](const auto& in_tile) mutable {
                            // Note: we can only selectively coalesce if all
                            // values inside the resulting larger tile will be
                            // values from the same root level.
                            const int  depth = in_tile.depth();
                            const auto index = in_tile.index();

                            // Compute tilesMaxChildDepth array
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                            if (depth == kMaxDepth - 1) {
                                // No need for separate check for maxDepth because
                                // we know depth == kMaxDepth - 1
                                if (!in_tile.isLeaf()) {
                                    tilesMaxChildDepth[depth][index] = 1;
                                }
                            } else {
                                for (int c = 0; c < Traits::kTileSize; ++c) {
                                    const auto childIndex =
                                        in_tile.childIndex(c);
                                    if (childIndex !=
                                        BIF::TileTree3D::kInvalidTileIndex) {
                                        tilesMaxChildDepth[depth][index] = std::max(tilesMaxChildDepth[depth][index], tilesMaxChildDepth[depth + 1][childIndex] + 1);
                                    }
                                }
                            }
#endif

                            if (!in_tile.hasLeaves()) {
                                if (depth == kMaxDepth - 1) {
                                    // In this case we know that all children at
                                    // kMaxDepth exist and we can
                                    // coalesce 8 4^3 tiles into a single 8^3
                                    // tile.
                                    tilesToCoalesce[depth][index] = 1;
                                } else {
                                    // In this case we must determine if all
                                    // children could also be recursively
                                    // coalesced from the same depth. That's the
                                    // case if tilesToCoalesce has the same
                                    // value for all children. We know that all
                                    // children exist.
                                    const int cd =
                                        tilesToCoalesce[depth + 1]
                                                       [in_tile.childIndex(0)];
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                                    // In this case we just need to check if all children
                                    // have the same path length. We do not need to ensure
                                    // that if the path length is zero, the children have to be leaves.
                                    // This allows for more coalescing.
                                    bool ok = (cd < Traits::kMaxCoalescedPath);
#else
                                    // In this case it's ok to coalesce if
                                    // - The path is smaller than the max allowed coalesced path.
                                    // - All children coalesced paths are equal and not zero
                                    //   or, all coalesced paths are zero and all children are leaves.
                                    const bool isZero = (cd == 0);
                                    bool       ok =
                                        (cd < Traits::kMaxCoalescedPath) &&
                                        ((isZero &&
                                          // We know child tile is at kMaxDepth - 1 or coarser.
                                          acc.tile(in_tile.childTreeIndex(0))
                                              .isLeaf()) ||
                                         !isZero);
#endif
                                    for (int c = 1; c < Traits::kTileSize && ok;
                                         ++c) {
                                        ok =
                                            (cd ==
                                                tilesToCoalesce[depth + 1]
                                                [in_tile.childIndex(
                                                    c)])
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                                            ;
#else
                                            && ((isZero &&
                                              // We know child tile is at kMaxDepth - 1 or coarser.
                                              acc.tile(
                                                     in_tile.childTreeIndex(c))
                                                  .isLeaf()) ||
                                             !isZero);
#endif
                                    }
                                    if (ok) {
                                        tilesToCoalesce[depth][index] = cd + 1;
                                    }
                                }

                                // Once we have processed whether this tile will be coalesced, we can determine
                                // if the children will be removed or not.
                                if (tilesToCoalesce[depth][index] > 0) {
                                    // In this case all children will be removed.
                                    for (int c = 0; c < Traits::kTileSize; ++c) {
                                        const auto childIndex = in_tile.childIndex(c);
                                        if (childIndex != BIF::TileTree3D::kInvalidTileIndex) {
                                            tilesToRemove[depth + 1][childIndex] = 1;
                                        }
                                    }
                                }
                            }
                        },
                            BIF::TileTree3D::TraversalMode::
                            kParallelBreadthFirstBottomUp,
                        BIF::TileTree3D::kInvalidTreeIndex, rootDepth,
                        kMaxDepth - 1);

#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                    // In this case we need an additional compaction step to determine proper indices
                    // when some tiles have been removed. We traverse from rootDepth + 1 because no tiles
                    // will be removed at the root depth.
                    {
                        // indexToVTT2 will contain 1, if the tile is kept. With a prefix sum
                        // we can then assign new indices.
                        Amino::Vector<int64_t> indexToVTT2[Traits::kMaxDepth + 1];
                        for (int d = rootDepth + 1; d <= kMaxDepth; ++d) {
                            // We add 1 additional entry for the prefix sum
                            indexToVTT2[d].resize(container->tileCount(d) + 1,
                                                  0);
                        }

                        // Map the permuted indices to a linearized space such that
                        // we can compute new compacted indices.
                        container->traverse(
                            [&](const auto& in_tile) {
                                const int     depth = in_tile.depth();
                                const int64_t index = in_tile.index();
                                if (!tilesToRemove[depth][index])
                                    indexToVTT2[depth][indexToVTT[depth][index] + 1] = 1;
                            },
                            BIF::TileTree3D::TraversalMode::kParallel,
                            BIF::TileTree3D::kInvalidTreeIndex, rootDepth + 1,
                            kMaxDepth);

                        // Compute the new compacted indices in the linearized space.
                        // \todo BIFROST-TBD here we could use a parallel prefix sum
                        for (int d = rootDepth + 1; d <= kMaxDepth; ++d) {
                            const uint64_t size = indexToVTT2[d].size();
                            for (uint64_t i = 1; i < size; ++i) {
                                indexToVTT2[d][i] += indexToVTT2[d][i - 1];
                            }
                        }

                        // Update the original mapping
                        container->traverse(
                            [&](const auto& in_tile) {
                                const int     depth = in_tile.depth();
                                const int64_t index = in_tile.index();
                                indexToVTT[depth][index] = static_cast<TileIndex>(indexToVTT2[depth][indexToVTT[depth][index]]);
                            },
                            BIF::TileTree3D::TraversalMode::kParallel,
                                BIF::TileTree3D::kInvalidTreeIndex, rootDepth + 1,
                                kMaxDepth);

                        for (int d = 0; d <= Traits::kMaxDepth + 1; ++d) {
                            if (d <= rootDepth) {
                                tileCountAccum[d] = 0;
                            }
                            else if (d == rootDepth + 1) {
                                tileCountAccum[d] = indexToVTT[d - 1].size();
                            }
                            else if (d <= kMaxDepth + 1) {
                                // Note that 'indexToVTT2[d][container->tileCount(d)]' is the number of tiles
                                // at depth 'd' after compaction.
                                tileCountAccum[d] =
                                    indexToVTT2[d - 1][container->tileCount(d - 1)] + tileCountAccum[d - 1];
                                if (verbose) {
                                    std::cout
                                        << "Compacted tiles[" << d - 1 << "] from "
                                        << container->tileCount(d - 1) << " to "
                                        << indexToVTT2[d - 1]
                                        [container->tileCount(d - 1)]
                                    << std::endl;
                                }
                            }
                            else {
                                tileCountAccum[d] = tileCountAccum[d - 1];
                            }
                        }
                    }
#endif // NANOVTT_SUPPORT_COALESCED_TILES
                } // if (coalesceTileData)
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                else
#endif // NANOVTT_SUPPORT_COALESCED_TILES
                {
                    for (int d = 0; d <= Traits::kMaxDepth + 1; ++d) {
                        if (d <= rootDepth)
                            tileCountAccum[d] = 0;
                        else
                            tileCountAccum[d] =
                            indexToVTT[d - 1].size() + tileCountAccum[d - 1];
                    }
                }

                auto tileCount = [&](const int in_depth) {
                    assert(in_depth >= rootDepth && in_depth <= Traits::kMaxDepth);
                    return tileCountAccum[in_depth + 1] - tileCountAccum[in_depth];
                };

                const float voxelSize = container->voxelSize();
                uint64_t containerSize = sizeof(Container);
                uint64_t containerSizeOpt = sizeof(Container);
                uint64_t numNanoVTTTiles = 0;
                uint64_t numVTTTiles = 0;
                for (int d = rootDepth; d <= kMaxDepth; ++d) {
                    numNanoVTTTiles += tileCount(d);
                    numVTTTiles += container->tileCount(d);
                    if (verbose && d < kMaxDepth)
                        containerSizeOpt += tileCount(d) * sizeof(uint64_t); // offset of tile from beginning of tile buffer
                } // for (int d = rootDepth; d <= kMaxDepth; ++d)

                Amino::Vector<uint64_t> tOffset;
                const uint64_t numNanoVTTTilesAtCoarserLevels =
                    numNanoVTTTiles -
                    tileCount(Traits::kMaxDepth);
                tbb::atomic<uint64_t> numLeafTilesAtCoarserLevels = 0;
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                tbb::atomic<uint64_t> tileStorageSize = 0;
#endif
                if (enableNanoVTTTraversal) {
                    int* traversalStructurePtr = nullptr;
                    size_t traversalStructureBufferSize =
                        numNanoVTTTiles * traversalStructureElementSize;
                    if (storeTileDataAs3DTextures) {
                        // The traversal structure will be stored as a 2D texture when the tile data is stored as 3D texture.
                        constexpr uint64_t rowSize = k2DTextureRowSize;
                        const uint64_t rowByteSize = rowSize * in_textureElementSize;
                        const uint64_t rowCount = traversalStructureBufferSize / rowByteSize + (traversalStructureBufferSize % rowByteSize == 0 ? 0 : 1);
                        traversalStructureBufferSize = rowByteSize * rowCount;
                    }
                    else if (in_padBufferForVolumetricTexture) {
                        unsigned int dim =
                            static_cast<unsigned int>(std::ceil(std::cbrt(static_cast<double>(traversalStructureBufferSize) / in_textureElementSize)));
                        traversalStructureBufferSize = dim * dim * dim * in_textureElementSize;
                    }
                    *out_traversalStructure = BufferT::create(
                        traversalStructureBufferSize, &in_allocator);
                    traversalStructurePtr = reinterpret_cast<int*>(
                        (*out_traversalStructure).data());

                    container->traverse(
                        [&](const auto& in_tile) {
                            const TileIndex accum = static_cast<TileIndex>(
                                tileCountAccum[in_tile.depth()]);
                            const TileIndex indexVTT =
                                accum +
                                indexToVTT[in_tile.depth()][in_tile.index()];
                            const size_t index =
                                indexVTT * traversalStructureElementCount;
                            traversalStructurePtr[index]     = in_tile.ti();
                            traversalStructurePtr[index + 1] = in_tile.tj();
                            traversalStructurePtr[index + 2] = in_tile.tk();
                            traversalStructurePtr[index + 3] =
                                in_tile.depth();
                        },
                        BIF::TileTree3D::TraversalMode::kParallel,
                        BIF::TileTree3D::kInvalidTreeIndex, rootDepth,
                        kMaxDepth);
                }

                if (enableNanoVTTNeighbors) {
                    auto acc = container->tileAccessor();
                    TileIndex* neighborStructurePtr = nullptr;
                    size_t neighborStructureBufferSize =
                        numNanoVTTTiles * neighborStructureElementSize;
                    if (storeTileDataAs3DTextures) {
                        // The neighbor structure will be stored as a 2D texture when the tile data is stored as 3D texture.
                        constexpr uint64_t rowSize = k2DTextureRowSize;
                        const uint64_t rowByteSize = rowSize * in_textureElementSize;
                        const uint64_t rowCount = neighborStructureBufferSize / rowByteSize + (neighborStructureBufferSize % rowByteSize == 0 ? 0 : 1);
                        neighborStructureBufferSize = rowByteSize * rowCount;
                    }
                    else if (in_padBufferForVolumetricTexture) {
                        unsigned int dim =
                            static_cast<unsigned int>(std::ceil(std::cbrt(static_cast<double>(neighborStructureBufferSize) / in_textureElementSize)));
                        neighborStructureBufferSize = dim * dim * dim * in_textureElementSize;
                    }
                    *out_neighborStructure = BufferT::create(
                        neighborStructureBufferSize, &in_allocator);
                    neighborStructurePtr = reinterpret_cast<TileIndex*>(
                        (*out_neighborStructure).data());

                    // \todo BIFROST-TBD - No need to compute structure if it already exists

                    container->traverse([&, acc](const auto& in_tile) mutable
                        // void beginTile(const Tile& in_tile)
                        {
                            const TileIndex accum = static_cast<TileIndex>(
                                tileCountAccum[in_tile.depth()]);
                            const TileIndex indexVTT =
                                accum +
                                indexToVTT[in_tile.depth()][in_tile.index()];
                            TileIndex* nbPtr = &neighborStructurePtr[27 * indexVTT];
                            for (int i = 0; i < 27; ++i, ++nbPtr) {
                                int oi = BIF::TileTree3D::NeighborOffset[i][0];
                                int oj = BIF::TileTree3D::NeighborOffset[i][1];
                                int ok = BIF::TileTree3D::NeighborOffset[i][2];
                                const BIF::TileTree3D::TreeIndex nbIndex = acc.tileNeighbor(in_tile, oi, oj, ok, in_tile.depth()).treeIndex();
                                if (nbIndex.m_depth >= rootDepth) {
                                    const TileIndex accum2 = static_cast<TileIndex>(
                                        tileCountAccum[nbIndex.m_depth]);
                                    const TileIndex indexVTT2 =
                                        accum2 +
                                        indexToVTT[nbIndex.m_depth][nbIndex.m_tile];
                                    *nbPtr = indexVTT2;
                                }
                                else {
                                    *nbPtr = kInvalidTileIndex;
                                }
                            } // for (int i = 0; i < 27; ++i, ++nbPtr)
                        }, BIF::TileTree3D::TraversalMode::kParallel,
                        BIF::TileTree3D::kInvalidTreeIndex, rootDepth,
                            kMaxDepth);
                } // if (enableNanoVTTNeighbors)

                {
                    const uint64_t tOffsetConstant = (numNanoVTTTilesAtCoarserLevels + 1) * sizeof(uint64_t) + numNanoVTTTiles * sizeof(TileIndex);
                    tOffset.resize(numNanoVTTTilesAtCoarserLevels + 1);
                    tOffset[0] = tOffsetConstant;
                    containerSize += tOffsetConstant;
                    // \todo BIFROST-TBD Use reduction instead
                    container->traverse(
                        [&](const auto& in_tile) {
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                            if (coalesceTileData && tilesToRemove[in_tile.depth()][in_tile.index()])
                                return;
#endif
                            const TileIndex accum =
                                static_cast<TileIndex>(tileCountAccum[in_tile.depth()]);
                            const TileIndex indexVTT =
                                accum +
                                indexToVTT[in_tile.depth()][in_tile.index()];

                            assert(static_cast<TileIndex>(in_tile.index()) < indexToVTT[in_tile.depth()].size());

#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                            if (coalesceTileData) {
                                const int coalescedPathLength =
                                    tilesToCoalesce[in_tile.depth()]
                                    [in_tile.index()];
                                const int maxChildDepth =
                                    tilesMaxChildDepth[in_tile.depth()]
                                    [in_tile.index()];
                                const bool isCoalesced =
                                    coalescedPathLength > 0 &&
                                    (in_tile.depth() == rootDepth ||
                                        tilesToCoalesce[in_tile.depth() - 1]
                                        [in_tile.parentIndex()] == 0);
                                const bool isCoalescedLeaf =
                                    (isCoalesced
                                        ? (coalescedPathLength == maxChildDepth)
                                        : false);
                                const bool isLeaf =
                                    in_tile.isLeaf() || isCoalescedLeaf || in_tile.depth() == kMaxDepth;
                                uint64_t sizeBytes = 0;
                                if (isLeaf) {
                                    // In the case of a leaf tile there's no need to store children
                                    ++numLeafTilesAtCoarserLevels;
                                }
                                else if (isCoalesced) {
                                    // If we have a coalesced tile that's not a leaf, then we need
                                    // to store children pointers. The coalescedPathLength determines how many.
                                    sizeBytes = sizeof(TileIndex) *
                                        static_cast<int>(std::pow(
                                            Traits::kTileWidth,
                                            // Note: we need to add 1 to the coalescedPathLength here to determine storage.
                                            (coalescedPathLength + 1) * 3));
                                }
                                else {
                                    sizeBytes = sizeof(TileIndex) * Traits::kTileSize;
                                }
                                tOffset[indexVTT + 1] = sizeBytes;
                                tileStorageSize += sizeBytes;
                            }
                            else {
                                if (in_tile.isLeaf() || in_tile.depth() == kMaxDepth) {
                                    ++numLeafTilesAtCoarserLevels;
                                    tOffset[indexVTT + 1] = 0;
                                }
                                else {
                                    tOffset[indexVTT + 1] = sizeof(TileIndex) * Traits::kTileSize;
                                    tileStorageSize += sizeof(TileIndex) * Traits::kTileSize;
                                }
                            }
#else
                            if (in_tile.isLeaf() || in_tile.depth() == kMaxDepth) {
                                ++numLeafTilesAtCoarserLevels;
                                tOffset[indexVTT + 1] = 0;
                            }
                            else {
                                tOffset[indexVTT + 1] = sizeof(TileIndex) * Traits::kTileSize;
                            }
#endif
                        },
                        BIF::TileTree3D::TraversalMode::kParallel,
                            BIF::TileTree3D::kInvalidTreeIndex, rootDepth,
                            std::min(kMaxDepth, Traits::kMaxDepth - 1));
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                    containerSize += tileStorageSize + sizeof(TileIndex) * Traits::kTileSize;
#else
                    containerSize += (numNanoVTTTilesAtCoarserLevels + 1 - numLeafTilesAtCoarserLevels) * sizeof(TileIndex) * Traits::kTileSize;
#endif
                    // \todo BIFROST-TBD here we could use a parallel prefix sum
                    for (uint64_t i = 1; i <= numNanoVTTTilesAtCoarserLevels; ++i) {
                        tOffset[i] += tOffset[i - 1];
                    }
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                    const uint64_t invalidTileOffset = tileStorageSize + tOffsetConstant;
#else
                    const uint64_t invalidTileOffset = (numNanoVTTTilesAtCoarserLevels - numLeafTilesAtCoarserLevels) * sizeof(TileIndex) * Traits::kTileSize + tOffsetConstant;
#endif
                    // \todo BIFROST-TBD Parallelize
                    for (uint64_t i = 0; i < numNanoVTTTilesAtCoarserLevels; ++i) {
                        if (tOffset[i] == tOffset[i + 1])
                            tOffset[i] = invalidTileOffset;
                    }
                }
                const uint64_t numNonLeavesAtCoarserDepths = numNanoVTTTilesAtCoarserLevels - numLeafTilesAtCoarserLevels;
                if (verbose) {
                    std::cout << "numNonLeavesAtCoarserDepths = "
                        << numNonLeavesAtCoarserDepths << "("
                        << static_cast<float>(numNonLeavesAtCoarserDepths) /
                        numNanoVTTTilesAtCoarserLevels * 100
                        << "%)" << std::endl;
            }
                if (verbose) {
                    std::cout << "Container size alignment = " << (containerSize % NANOVDB_DATA_ALIGNMENT) << std::endl;
                }
                // Make sure following data is aligned
                containerSize += NANOVDB_DATA_ALIGNMENT - (containerSize % NANOVDB_DATA_ALIGNMENT);

            Amino::Vector<uint64_t> tdOffset;
            Amino::Vector<uint64_t> tdcOffset;
            Amino::Vector<Amino::Vector<uint64_t>> tdOffsetFpNVec;
            uint64_t fpNIndex = 0;
            // Find the depth at which we need to optimize to select the tiles for which we should build a halo.
            // Note that the budget is found from the tiles in the VTT, not the possibly coalesced NanoVTT.
            int haloOptDepth = rootDepth + 1;
            float haloOptFraction = 0.f;
            for (; haloOptDepth <= kMaxDepth && haloOptFraction < in_extendedBudgetFraction; ++haloOptDepth) {
                haloOptFraction = static_cast<float>(tileCountAccum[haloOptDepth + 1]) / numVTTTiles;
            }
            haloOptDepth = std::max(rootDepth + 1, haloOptDepth - 1);
            haloOptFraction = std::min(static_cast<float>(tileCountAccum[haloOptDepth]) / numVTTTiles, in_extendedBudgetFraction);
            const TileIndex haloOptBudget = static_cast<TileIndex>((in_extendedBudgetFraction - haloOptFraction) * numVTTTiles);
            if (verbose) {
                std::cout << "haloOptDepth = " << haloOptDepth << ", haloOptFraction = " << haloOptFraction << ", haloOptBudget = " << haloOptBudget << std::endl;
            }

            // Determine for which tiles we need a halo
            tdOffset.resize(numNanoVTTTiles + 1, 0);
            tdOffset[0] = 0;
            if (coalesceTileData) {
                tdcOffset.resize(numNanoVTTTiles + 1, 0);
                tdcOffset[0] = 0;
            }

            tbb::atomic<uint64_t> numHaloTiles = 0;
            tbb::atomic<uint64_t> numNonHaloTiles = 0;

            auto setOffset = [&](const bool in_needsHalo, const TileIndex in_indexVTT)
            {
                if (!in_needsHalo) {
                    // Was already initialized to zero
                    ++numNonHaloTiles;
                }
                else {
                    tdOffset[in_indexVTT + 1] = 1;
                    ++numHaloTiles;
                }
            };

            auto acc = container->tileAccessor();

            // The halos generated using the optimization approach are only 1% faster at render time,
            // but adds 26% to the conversion time.
            const bool generateHalosUsingOptimization = false;

            if (!coalesceTileData && in_extendedBudgetFraction <= 0)
            {
                numNonHaloTiles = numNanoVTTTiles;
                numHaloTiles = 0;
            }
            else
            {
                if (!generateHalosUsingOptimization) {
                    // \todo BIFROST-TBD Use reduction instead.
                    container->traverse(
                        [&, acc](const auto& in_tile) {
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                            if (coalesceTileData && tilesToRemove[in_tile.depth()][in_tile.index()])
                                return;
#endif
                            const TileIndex accum =
                                static_cast<TileIndex>(tileCountAccum[in_tile.depth()]);
                            const TileIndex indexVTT =
                                accum +
                                indexToVTT[in_tile.depth()][in_tile.index()];
                            bool needsHalo = false;
                            if (in_extendedBudgetFraction > 0) {
                                const bool isAtRootDepth =
                                    (in_tile.depth() == rootDepth);
                                needsHalo = isAtRootDepth;
                                if (!needsHalo) {
                                    int        numHaloPoints = 0;
                                    const bool graded = checkGraded(
                                        in_tile, acc, numHaloPoints);
                                    const bool isAtResJump =
                                        (numHaloPoints > 0);
                                    const bool hasLeaves = (in_tile.hasLeaves() || in_tile.depth() == kMaxDepth);
                                    const bool isRequiredAtCoarserLevel =
                                        (!graded || hasLeaves || isAtResJump || (coalesceTileData && in_extendedBudgetFraction >= 1));
                                    needsHalo =
                                        (((in_tile.depth() < haloOptDepth) ||
                                            (in_tile.depth() == haloOptDepth &&
                                                in_tile.index() < haloOptBudget)) &&
                                            isRequiredAtCoarserLevel) ||
                                        !graded || (in_extendedBudgetFraction >= 1);
                                }
                            } // if (in_extendedBudgetFraction > 0)
                            setOffset(needsHalo, indexVTT);
                            if (coalesceTileData) {
                                const bool tileIsCoalesced =
                                    !(in_tile.depth()==kMaxDepth) &&
                                    tilesToCoalesce[in_tile.depth()]
                                    [in_tile.index()] > 0 &&
                                    (in_tile.depth() == rootDepth ||
                                        tilesToCoalesce[in_tile.depth() - 1]
                                        [in_tile.parentIndex()] == 0);
                                const bool tileIsRemoved =
                                    tilesToRemove[in_tile.depth()]
                                                 [in_tile.index()];
                                const int cp =
                                    (tileIsCoalesced
                                        ? tilesToCoalesce[in_tile.depth()]
                                        [in_tile.index()]
                                : 0);
                                const int coalescedTileDataSize =
                                    (tileIsRemoved
                                        ? 0
                                        : static_cast<int>(std::pow(
                                            Traits::kTileDataWidth *
                                            std::pow(Traits::kTileWidth,
                                                cp) +
                                            (needsHalo ? 1 : 0),
                                            3)));
                                tdcOffset[indexVTT + 1] = coalescedTileDataSize;
                        }
                        },
                        BIF::TileTree3D::TraversalMode::kParallel,
                        BIF::TileTree3D::kInvalidTreeIndex, rootDepth,
                        kMaxDepth);
                } // if (!generateHalosUsingOptimization)
                else {
                    // \todo BIFROST-TBD not supported
                    assert(false);
                    // First traverse upper levels where all tiles will be halos.
                    container->traverse(
                        [&, acc](const auto& in_tile) {
                            const TileIndex accum =
								static_cast<TileIndex>(tileCountAccum[in_tile.depth()]);
                            const TileIndex indexVTT =
                                accum +
                                indexToVTT[in_tile.depth()][in_tile.index()];
                            tdOffset[indexVTT + 1] = 1;
                            ++numHaloTiles;
                        },
                        BIF::TileTree3D::TraversalMode::kParallel,
                            BIF::TileTree3D::kInvalidTreeIndex, rootDepth,
                            haloOptDepth - 1);

                    // Second, traverse the haloOptDepth to find out which tiles
                    // should be halos.
                    if (haloOptDepth > rootDepth) {
                        const auto numTiles = tileCountAccum[haloOptDepth + 1] -
                            tileCountAccum[haloOptDepth];
                        Amino::Vector<int>                haloOptVec(numTiles);
                        Amino::Vector<TileIndex> haloOptIndexVec(numTiles);
                        Amino::Vector<TileIndex> haloOptIndexVec2(
                            numTiles);
                        // First we determine the number of halo points needed for
                        // each tile
                        container->traverse(
                            [&, acc](const auto& in_tile) {
                                int        numHaloPoints = 0;
                                const bool graded =
                                    checkGraded(in_tile, acc, numHaloPoints);
                                haloOptVec[in_tile.index()] =
                                    (graded ? numHaloPoints
                                        : std::numeric_limits<int>::max());
                                haloOptIndexVec[in_tile.index()] = static_cast<TileIndex>(in_tile.index());
                            },
                            BIF::TileTree3D::TraversalMode::kParallel,
                                BIF::TileTree3D::kInvalidTreeIndex, haloOptDepth,
                                haloOptDepth);
                        // Second we need to determine the haloOptBudget-largest
                        // element For now, we sort in descending order.
                        auto compare = [&](const auto& in1,
                            const auto& in2) -> bool {
                                return haloOptVec[in1] > haloOptVec[in2];
                        };
                        tbb::parallel_sort(haloOptIndexVec.begin(),
                            haloOptIndexVec.end(), compare);
                        // Third we create the inverse mapping
                        tbb::parallel_for(
                            tbb::blocked_range<TileIndex>(0, static_cast<TileIndex>(numTiles)),
                            [&](const tbb::blocked_range<TileIndex>&
                                in_r) {
                                    for (auto i = in_r.begin(); i != in_r.end(); ++i) {
                                        haloOptIndexVec2[haloOptIndexVec[i]] = i;
                                    }
                            });
                        // Finally we determine which tiles are halos and which are
                        // not
                        container->traverse(
                            [&, acc](const auto& in_tile) {
                                const TileIndex accum =
									static_cast<TileIndex>(tileCountAccum[in_tile.depth()]);
                                const TileIndex indexVTT =
                                    accum +
                                    indexToVTT[in_tile.depth()][in_tile.index()];
                                const bool needsHalo =
                                    (haloOptVec[in_tile.index()] ==
                                        std::numeric_limits<int>::max() ||
                                        haloOptIndexVec2[in_tile.index()] <
                                        haloOptBudget);

                                setOffset(needsHalo, indexVTT);
                            },
                            BIF::TileTree3D::TraversalMode::kParallel,
                                BIF::TileTree3D::kInvalidTreeIndex,
                                haloOptDepth, haloOptDepth);
                    } // if (haloOptDepth > rootDepth)

                    // Finally traverse the finer levels where only non-graded tiles
                    // will be halos.
                    if (haloOptDepth < kMaxDepth) {
                        container->traverse(
                            [&, acc](const auto& in_tile) {
                                const TileIndex accum =
									static_cast<TileIndex>(tileCountAccum[in_tile.depth()]);
                                const TileIndex indexVTT =
                                    accum +
                                    indexToVTT[in_tile.depth()][in_tile.index()];
                                int        numHaloPoints = 0;
                                const bool graded =
                                    checkGraded(in_tile, acc, numHaloPoints);
                                setOffset(!graded, indexVTT);
                            },
                            BIF::TileTree3D::TraversalMode::kParallel,
                                BIF::TileTree3D::kInvalidTreeIndex, haloOptDepth + 1,
                                kMaxDepth);
                    } // if (haloOptDepth < kMaxDepth)
                }
            }

            if (verbose)
            {
                std::cout << "numNanoVTTTiles = " << numNanoVTTTiles << std::endl;
                std::cout << "numHaloTiles = " << numHaloTiles << " (" << static_cast<float>(numHaloTiles) / (numHaloTiles + numNonHaloTiles) * 100 << "%)" << std::endl;
                std::cout << "numNonHaloTiles = " << numNonHaloTiles << std::endl;
            }
            // \todo BIFROST-TBD Here we could use a parallel prefix sum
            for (uint64_t i = 1; i <= numNanoVTTTiles; ++i) {
                tdOffset[i] += tdOffset[i - 1];
                if (coalesceTileData)
                    tdcOffset[i] += tdcOffset[i - 1];
            }
            if (verbose) {
                std::cout << "tdOffset[numNanoVTTTiles] = " << tdOffset[numNanoVTTTiles] << std::endl;
            }

            Amino::Vector<uint64_t> gridAndTreeSize;
            uint64_t gridAndTreeSizeTotal = 0;


            // DEBUG INFO BEGIN

            // Determine how close to Morton storage the input is
            if (outputMortonOrderStorageInfo) {
                BIF::TileTree3D::MortonCode mcPrev;
                int64_t                     numMortonChecks = 0;
                int64_t                     numMortonOK = 0;
                int64_t                     numMortonNotOK = 0;
                container->traverse(
                    [&](const auto& in_tile) {
                        BIF::TileTree3D::MortonCode mc =
                            BIF::TileTree3D::computeMortonCode<Traits>(
                                in_tile.ti(), in_tile.tj(), in_tile.tk());
                        if (numMortonChecks > 0) {
                            if (!(mc < mcPrev)) {
                                ++numMortonOK;
                            }
                            else {
                                ++numMortonNotOK;
                            }
                        }
                        mcPrev = mc;
                        ++numMortonChecks;
                    },
                    BIF::TileTree3D::TraversalMode::kBreadthFirst,
                        BIF::TileTree3D::kInvalidTreeIndex, 0, kMaxDepth);
                std::cout << "Morton stats: numOK = " << numMortonOK
                    << ", numNotOK = " << numMortonNotOK << std::endl;
            }

            // Determine how much storage required for 8^3 tiles
            if (outputTraits28StorageInfo)
            {
                using Traits8 = BIF::TileTree3D::Traits<2, 8>;
                Amino::Vector<Bifrost::Math::int3> tilesTraits8[Traits::kMaxDepth + 1];
                container->traverse(
                    [&](const auto& in_tile) {
                        const auto tijk  = in_tile.tijk();
                        const int  depth = in_tile.depth();
                        // Determine which tile tijk maps to with Traits8
                        const int index8 = Traits8::kMaxDepth -
                                           (Traits::kMaxDepth - depth);
                        const int        tvw8 = Traits8::kTileSideLUT[index8];
                        const Bifrost::Math::int3 tijk8{
                            (tijk.x - Traits8::kMinTileCoord) / tvw8 * tvw8 +
                                Traits8::kMinTileCoord,
                            (tijk.y - Traits8::kMinTileCoord) / tvw8 * tvw8 +
                                Traits8::kMinTileCoord,
                            (tijk.z - Traits8::kMinTileCoord) / tvw8 * tvw8 +
                                Traits8::kMinTileCoord};
                        tilesTraits8[depth].push_back(tijk8);
                    },
                    BIF::TileTree3D::TraversalMode::kBreadthFirst,
                        BIF::TileTree3D::kInvalidTreeIndex, rootDepth, Traits::kMaxDepth);

                // Now sort and then count
                std::cout << "Traits8 storage requirements" << std::endl;
                uint64_t numTilesTraits8 = 0;
                for (int depth = rootDepth; depth <= Traits::kMaxDepth; ++depth) {
                    tbb::parallel_sort(tilesTraits8[depth].begin(), tilesTraits8[depth].end());
                    uint64_t numTiles = 0;
                    const size_t size = tilesTraits8[depth].size();
                    for (size_t i = 0; i < size; ++i) {
                        if (i == size || tilesTraits8[depth][i + 1] != tilesTraits8[depth][i])
                            ++numTiles;
                    }
                    numTilesTraits8 += numTiles;
                    std::cout << " numTiles[" << Traits8::kMaxDepth - (Traits::kMaxDepth - depth) << "] = "
                        << numTiles << " (Traits28) vs numTiles[" << depth << "] = " << tileCount(depth) << " (Traits24)" << std::endl;
                }
                std::cout << " numTiles total = " << numTilesTraits8 << " (Traits28) vs " << numNanoVTTTiles << " (Traits24)" << std::endl;
                std::cout << " 0% extended tile data storage = "
                          << static_cast<float>(numTilesTraits8 *
                                                Traits8::kTileDataSize *
                                                sizeof(float)) /
                                 1024 / 1024
                          << "MB (Traits28) vs "
                          << static_cast<float>(numNanoVTTTiles *
                                                Traits::kTileDataSize *
                                                sizeof(float)) /
                                 1024 / 1024
                          << "MB (Traits24)" << std::endl;
                std::cout << " 100% extended tile data storage = "
                          << static_cast<float>(
                                 numTilesTraits8 *
                                 (Traits8::kTileDataWidth + 1) *
                                 (Traits8::kTileDataWidth + 1) *
                                 (Traits8::kTileDataWidth + 1) * sizeof(float)) /
                                 1024 / 1024
                          << "MB (Traits28) vs "
                          << static_cast<float>(
                                 numNanoVTTTiles *
                                 (Traits::kTileDataWidth + 1) *
                                 (Traits::kTileDataWidth + 1) *
                                 (Traits::kTileDataWidth + 1) * sizeof(float)) /
                                 1024 / 1024
                          << "MB (Traits24)" << std::endl;
            }

            if (outputTileDataCoarseningStorageInfo) {

                Amino::Vector<int> tilesToCoarsen[Traits::kMaxDepth];
                for (int d = rootDepth; d < Traits::kMaxDepth; ++d) {
                    tilesToCoarsen[d].resize(container->tileCount(d), 0);
                }
                uint64_t numTilesToCoarsen = 0;
                int maxCoarsenedDepth = 0;
                int minDepthForCoarsening = Traits::kMaxDepth;
                int maxDepthForCoarsening = 0;
                container->traverse(
                    [&, acc](const auto& in_tile) {
                        // Note: we can only selectively coarsen if all values inside the resulting
                        // larger tile will be values from the same root level.
                        const int depth = in_tile.depth();
                        const auto index = in_tile.index();
                        if (!in_tile.hasLeaves()) {
                            if (depth == Traits::kMaxDepth - 1) {
                                // In this case we know that all children at Traits::kMaxDepth exist
                                // and we can coarsen 8 4^3 tiles into a single 8^3 tile.
                                numTilesToCoarsen += Traits::kTileSize;
                                tilesToCoarsen[depth][index] = 1;
                                maxCoarsenedDepth = std::max(maxCoarsenedDepth, 1);
                                minDepthForCoarsening = std::min(depth, minDepthForCoarsening);
                                maxDepthForCoarsening = std::max(depth, maxDepthForCoarsening);
                            }
                            else {
                                // In this case we must determine if all children could also be recursively
                                // coarsened from the same depth. That's the case if tilesToCoarsen has the same value for all children.
                                const int cd = tilesToCoarsen[depth + 1][in_tile.childIndex(0)];
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                                // In this case we just need to check if all children
                                // have the same path length. We do not need to ensure
                                // that if the path length is zero, the children have to be leaves.
                                // This allows for more coalescing.
                                bool ok = (cd < Traits::kMaxCoalescedPath);
#else
                                // In this case it's ok to coalesce if
                                // - The path is smaller than the max allowed coalesced path.
                                // - All children coalesced paths are equal and not zero
                                //   or, all coalesced paths are zero and all children are leaves.
                                const bool isZero = (cd == 0);
                                bool       ok =
                                    (cd < Traits::kMaxCoalescedPath) &&
                                    ((isZero &&
                                        acc.tile(in_tile.childTreeIndex(0))
                                        .isLeaf()) ||
                                        !isZero);
#endif
                                for (int c = 1; c < Traits::kTileSize && ok;
                                    ++c) {
                                    ok =
                                        (cd ==
                                            tilesToCoarsen[depth + 1]
                                            [in_tile.childIndex(
                                                c)])
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                                        ;
#else
                                        && ((isZero &&
                                            acc.tile(
                                                in_tile.childTreeIndex(c))
                                            .isLeaf()) ||
                                            !isZero);
#endif
                                }
                                if (ok) {
                                    numTilesToCoarsen += Traits::kTileSize;
                                    tilesToCoarsen[depth][index] = cd + 1;
                                    maxCoarsenedDepth = std::max(maxCoarsenedDepth, cd + 1);
                                    minDepthForCoarsening = std::min(depth, minDepthForCoarsening);
                                    maxDepthForCoarsening = std::max(depth, maxDepthForCoarsening);
                                }
                            }
                        }
                    },
                    BIF::TileTree3D::TraversalMode::kBreadthFirstBottomUp,
                        BIF::TileTree3D::kInvalidTreeIndex, rootDepth, Traits::kMaxDepth - 1);

                std::cout << "Selective coarsening storage requirements" << std::endl;
                std::cout << " numTilesToCoarsen = " << numTilesToCoarsen << " ("
                          << static_cast<float>(numTilesToCoarsen) /
                                 numNanoVTTTiles * 100
                          << "%)" << std::endl;
                std::cout << " maxCoarsenedDepth = " << maxCoarsenedDepth
                          << " corresponding to a tiledata of size "
                          << static_cast<float>(Traits::kTileDataWidth) * std::pow(static_cast<float>(Traits::kTileWidth),
                                      static_cast<float>(maxCoarsenedDepth))
                          << "^3" << std::endl;
                std::cout << " minDepthForCoarsening = " << minDepthForCoarsening
                          << ", maxDepthForCoarsening = "
                          << maxDepthForCoarsening << std::endl;
                if (maxCoarsenedDepth == 1) {
                    // Note: these numbers are only correct if maxCoarsenedDepth
                    // == 1: memory counts the children coarsened as well as the
                    // single parent they're coarsened into.
                    const uint64_t memBefore =
                        (numTilesToCoarsen + numTilesToCoarsen / 8) *
                        Traits::kExtendedTileDataSize * sizeof(float);
                    const uint64_t memAfter =
                        (numTilesToCoarsen / 8) *
                        (Traits::kTileDataSize * Traits::kTileSize + 1) *
                        sizeof(float);
                    std::cout << " memory savings for 100% extended tile data storage: = "
                              << static_cast<float>(memBefore - memAfter) /
                                     1024 / 1024
                              << "MB" << std::endl;
                }
            }
            // DEBUG INFO END
            endTime2 = tbb::tick_count::now();
            if (verbose || verboseTimings) {
                std::cout << "Container pass 1 = "
                    << (endTime2 - startTime2).seconds() << " seconds" << std::endl;
            }

            // CONTAINER.END (pass 1)

            // CHANNELS.BEGIN (pass 1)
            startTime2 = tbb::tick_count::now();

            uint32_t gridCount = 0;
            Amino::Vector<Amino::Vector<Codec>> codecsVec;
            uint64_t tdOffsetConstant = (numNanoVTTTiles + 1) * sizeof(uint64_t);
            OracleT oracle(in_tolerance);
            nanovdb::DitherLUT lut(in_ditherOn);

            BIF::Utils::ChannelParser parser;
            parser.parses(in_properties);

            // If the tile data is stored as a 3D texture, the
            // data for all properties must use the same dimensions.
            uint64_t texture3DDim = 0;
            bool texture3DDimInitialized = false;

            Amino::for_each<NanoVTTTypes>([&](auto in) {
                using PairType = decltype(in);
                using VTTType = std::tuple_element_t<0, PairType>;
                using NanoVTTType =
                    std::tuple_element_t<1, PairType>;
                using BuildType = typename std::conditional<
                    std::is_same<NanoVTTType, float>::value && !std::is_same<VoxelT, void>::value,
                    VoxelT, NanoVTTType>::type;
                constexpr uint64_t valueTypeSize = sizeof(typename Tree<BuildType>::ValueType);
                using BIFChannel =
                    BIF::TileTree3D::Channel<BIFTraits, VTTType,
                                             BIFTraits::kTileDataWidth>;

                for (const Amino::String& name : in_volume.componentRange()) {
                    if (!parser.accepts(name)) continue;
                    if (auto channel = in_volume.component<BIFChannel>(name)) {
                        tbb::atomic<uint64_t> treeSize = 0;

                        // Here we need to compute the size explicitly for variable bitWidth quantization
                        if (std::is_same<BuildType, nanovdb::FpN>::value) {
                            treeSize = numNanoVTTTiles * sizeof(uint64_t);
                            codecsVec.resize(codecsVec.size() + 1);
                            auto& codecs = codecsVec.back();
                            codecs.resize(numNanoVTTTiles);
                            tdOffsetFpNVec.resize(tdOffsetFpNVec.size() + 1);
                            auto& tdOffsetFpN = tdOffsetFpNVec.back();
                            tdOffsetFpN.resize(numNanoVTTTiles);
                            nanovdb::GridClass gridClass = nanovdb::GridClass::Unknown;
                            const bool isSDF = ::BIF::TileTree3D::LevelSet::isSignedDistance(*channel);
                            if (isSDF) {
                                gridClass = nanovdb::GridClass::LevelSet;
                            }
                            else if (name == Bifrost::Geometry::sVoxelFogDensity) {
                                gridClass = nanovdb::GridClass::FogVolume;
                            }
                            initOracle<BuildType>(gridClass,
                                voxelSize,
                                oracle);
                            if (verbose)
                                std::cout << "tolerance = " << oracle.getTolerance() << std::endl;
                            // Since we're adding integers we can use an atomic for the reduction.
                            // \todo BIFROST-TBD Use reduction here instead.
                            tbb::parallel_for(
                                tbb::blocked_range<int>(rootDepth,
                                                        kMaxDepth + 1,
                                                        /*grainsize*/ 1),
                                [&](const tbb::blocked_range<int>& in_rd) {
                                    for (int d = in_rd.begin();
                                         d != in_rd.end(); ++d) {
                                        const int64_t size =
                                            container->tileCount(d);
                                        const int64_t accum = tileCountAccum[d];
                                        tbb::parallel_for(
                                            tbb::blocked_range<int64_t>(0,
                                                                        size),
                                            [&](const tbb::blocked_range<
                                                int64_t>& in_r) {
                                                uint64_t treeSizeLocal = 0;
                                                for (auto i = in_r.begin();
                                                     i != in_r.end(); ++i) {
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                                                    if (coalesceTileData && tilesToRemove[d][i])
                                                        continue;
#endif
                                                    const int64_t indexVTT =
                                                        accum +
                                                        indexToVTT[d][i];
                                                    compression<BuildType>(
                                                        &channel->tileData(
                                                            {i, d})(0),
                                                        lut, oracle,
                                                        codecs[indexVTT]
                                                        // Note that for halo
                                                        // tiles, the halo
                                                        // itself is not
                                                        // included here but we
                                                        // assume a smooth
                                                        // transition between
                                                        // tiles.
                                                        ,
                                                        Traits::kTileDataSize,
                                                        ((tdOffset[indexVTT +
                                                                   1] -
                                                          tdOffset[indexVTT]) ==
                                                                 0
                                                             ? Traits::
                                                                   kTileDataSize
                                                             : Traits::
                                                                   kExtendedTileDataSize));
                                                    treeSizeLocal +=
                                                        codecs[indexVTT].size +
                                                        Tree<BuildType>::
                                                            kTileDataBytes;
                                                }
                                                treeSize += treeSizeLocal;
                                            });
                                    }
                                });
                            // \todo BIFROST-TBD Could be optimized by a parallel prefix sum
                            // Note: in this case since FpN is only supported for floats
                            // we can fully compute and re-use the offset table here.
                            tdOffsetFpN[0] = numNanoVTTTiles * sizeof(uint64_t);
                            for (uint64_t i = 1; i < numNanoVTTTiles; ++i) {
                                tdOffsetFpN[i] = codecs[i-1].size + Tree<BuildType>::kTileDataBytes + tdOffsetFpN[i-1];
                            }
                        } // if (std::is_same<BuildType, nanovdb::FpN>::value)
                        else {
                            // Note: since the table is re-used for multiple types we need to account for the
                            //       size of the type later.
                            if (std::is_same<VoxelT, void>::value || !std::is_same<VTTType, float>::value) {
                                if (coalesceTileData) {
                                    treeSize = tdOffsetConstant + tdcOffset[numNanoVTTTiles] * valueTypeSize;
                                }
                                else if (!storeTileDataAs3DTextures)
                                {
                                    treeSize = numHaloTiles * valueTypeSize *
                                        Traits::kExtendedTileDataSize +
                                        numNonHaloTiles * valueTypeSize *
                                        Traits::kTileDataSize +
                                        tdOffsetConstant;
                                }
                                else {
                                    // storeTileDataAs3DTextures is true
                                    assert(numNonHaloTiles == 0);
                                    const uint64_t dim = static_cast<
                                        unsigned int>(std::ceil(std::cbrt(
                                            // Note: we add 1 to account for a tile of default values
                                        static_cast<double>(numHaloTiles + 1))) * Traits::kExtendedTileDataWidth);
                                    treeSize =
                                        dim * dim * dim *
                                        sizeof(typename Tree<BuildType>::ValueType) +
                                        tdOffsetConstant;
                                    if (verbose)
                                    {
                                        std::cout << "texture dim = " << dim
                                            << " for property "
                                            << name.c_str()
                                            << std::endl;
                                    }

                                    if (!texture3DDimInitialized) {
                                        texture3DDim = dim;
                                    }
                                    assert(dim == texture3DDim);
                                }
                            }
                            else {
                                // Fp4, Fp8, Fp16
                                treeSize =
                                    numHaloTiles *
                                    Tree<BuildType>::kExtendedTileDataBytes +
                                    numNonHaloTiles *
                                    Tree<BuildType>::kTileDataBytes +
                                    tdOffsetConstant;
                                if (verbose) {
                                    std::cout << "treeSize (quantization) = " << treeSize << std::endl;
                                }
                            }
                        }
                        gridAndTreeSize.push_back(
                            sizeof(Tree<BuildType>) +
                            sizeof(Grid<Tree<BuildType>>) +
                            treeSize
                        );
                        // Make sure following data is aligned
                        gridAndTreeSize.back() += NANOVDB_DATA_ALIGNMENT - (gridAndTreeSize.back() % NANOVDB_DATA_ALIGNMENT);

                        gridAndTreeSizeTotal += gridAndTreeSize.back();
                        if (verbose) {
                            std::cout << "Grid #" << gridCount << ": Found channel of type '"
                                << typeid(VTTType).name() << "' and name '"
                                << name.c_str() << "' and grid size "
                                << gridAndTreeSize.back() << std::endl;
                        }
                        ++gridCount;
                    }
                }
            });

            if (verbose) {
                std::cout << "Found #" << gridCount << " number of grids"
                    << std::endl;
            }
            endTime2 = tbb::tick_count::now();
            if (verbose || verboseTimings) {
                std::cout << "Channels pass 1 = "
                    << (endTime2 - startTime2).seconds() << " seconds" << std::endl;
            }

            // CHANNELS.END (pass 1)

            // Second pass: Here we allocate and compute the buffer
            startTime2 = tbb::tick_count::now();

            uint64_t bufferSize = containerSize + gridAndTreeSizeTotal;
            bufferSize += NANOVDB_DATA_ALIGNMENT - (bufferSize % NANOVDB_DATA_ALIGNMENT);

            if (verbose)
            {
                std::cout << "numNanoVTTTiles = " << numNanoVTTTiles << std::endl;
                std::cout << "bufferSize = " << bufferSize << std::endl;
                std::cout << " - Size of TileDataVariable = " << sizeof(TileDataVariable) << std::endl;
                std::cout << " - Size of TreeIndex = "
                          << sizeof(TreeIndex) << std::endl;
                std::cout << " - Size of VoxelCoord = "
                    << sizeof(VoxelCoord) << std::endl;
                std::cout << " - Size of TileIndex = " << sizeof(TileIndex)
                          << std::endl;
                std::cout << " - Size of nanovdb::Coord = "
                          << sizeof(nanovdb::Coord) << std::endl;
                std::cout << " - Traits::kTileSize = "
                          << Traits::kTileSize << std::endl;
                std::cout << "containerSize = " << containerSize << "(" << static_cast<float>(containerSize)/bufferSize*100 << "%)" << std::endl;
                std::cout << "gridAndTreeSizeTotal = " << gridAndTreeSizeTotal << "(" << static_cast<float>(gridAndTreeSizeTotal) / bufferSize * 100 << "%)"
                          << std::endl;
            }

            if (in_padBufferForVolumetricTexture) {
                uint64_t dim =
                    static_cast<unsigned int>(std::ceil(std::cbrt(static_cast<double>(bufferSize) / in_textureElementSize)));
                bufferSize = dim * dim * dim * in_textureElementSize;
            }
            else {
                // The container etc will be stored as 2D texture with a fixed row size so we make sure the whole
                // buffer is aligned with that size.
                constexpr uint64_t rowSize = k2DTextureRowSize;
                const uint64_t rowByteSize = rowSize * in_textureElementSize;
                const uint64_t rowCount = bufferSize / rowByteSize + (bufferSize % rowByteSize == 0 ? 0 : 1);
                bufferSize = rowByteSize * rowCount;
            }

            // To save as an nvtt file, the first 8 bytes contain the
            // size of the file
            handle = BufferT::create(bufferSize, &in_allocator);
            // Make sure buffer contains no uninitialized memory which can otherwise cause uninitialized
            // memory in data required for alignment.
            std::memset(handle.data(), 0, bufferSize);

            endTime2 = tbb::tick_count::now();
            if (verbose || verboseTimings) {
                std::cout << "Allocation and clearing buffers = "
                    << (endTime2 - startTime2).seconds() << " seconds" << std::endl;
            }


            // CONTAINER.BEGIN (pass 2)
            startTime2 = tbb::tick_count::now();

            // Container
            Container* containerVTT =
                reinterpret_cast<Container*>(handle.data());
            containerVTT->m_bufferSize = bufferSize;
            containerVTT->m_containerSize     = containerSize;
            containerVTT->m_parentOffset = (numNanoVTTTilesAtCoarserLevels + 1) * sizeof(uint64_t);
            for (int d = 0; d <= Traits::kMaxDepth + 1; ++d) {
                containerVTT->m_tileCountAccum[d] = static_cast<TileIndex>(tileCountAccum[d]);
            }
            {
                // Set up the hash table
                auto acc2                  = container->tileAccessor();
                containerVTT->m_rootDepth = rootDepth;
                for (uint32_t i = 0; i < Container::kTableSize; ++i) {
                    containerVTT->m_table[i] =
                        kInvalidTileIndex;
                }
                containerVTT->m_tableMin = hashTableMin;
                for (int64_t i = 0; i < tileCount(rootDepth); ++i) {
                    auto& tile = acc2.tile(i, rootDepth);
                    containerVTT->addRoot(
                        {tile.ti(), tile.tj(), tile.tk()},
                        static_cast<TileIndex>(in_sortIntoMortonOrder
                                                   ? permutation[rootDepth][i]
                                                   : i));
                }
            }

            if (verbose) {
                std::cout << "voxelSize = " << voxelSize << std::endl;
                for (int d = 0; d <= Traits::kMaxDepth + 1; ++d) {
                    std::cout << "tileCountAccum[" << d
                              << "] = " << containerVTT->m_tileCountAccum[d]
                              << std::endl;
                }
                for (int d = 0; d <= Traits::kMaxDepth; ++d) {
                    std::cout << "tileCount[" << d << "] = "
                              << containerVTT->m_tileCountAccum[d + 1] -
                                     containerVTT->m_tileCountAccum[d]
                              << std::endl;
                }
                std::cout << "BufferSize = " << bufferSize << std::endl;
            }

            // \todo BIFROST-TBD - We could fuse this loop with the loop over tile data and then only
            //                     initialize the container for the first loop over tile data.

            tbb::atomic<uint64_t> numTilesCoalesced = 0;

            container->traverse(
                [&, acc](const auto& in_tile) mutable {
                    const int     depth = in_tile.depth();
                    // Note that in this case a coalesced tile is being terminated because all
                    // children have been moved into that coalesced tile.
                    const bool tileIsCoalesced = coalesceTileData &&
                        !(in_tile.depth()==kMaxDepth) &&
                        tilesToCoalesce[depth][in_tile.index()] > 0 &&
                        (depth == rootDepth ||
                         tilesToCoalesce[depth - 1][in_tile.parentIndex()] ==
                             0);
#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                    const bool tileIsRemoved = (coalesceTileData ?
                        tilesToRemove[in_tile.depth()]
                        [in_tile.index()] : false);
                    const int coalescedPathLength = (tileIsCoalesced ?
                        tilesToCoalesce[in_tile.depth()][in_tile.index()] : 0);
                    const int maxChildDepth = (tileIsCoalesced ?
                        tilesMaxChildDepth[in_tile.depth()]
                        [in_tile.index()] : 0);
                    const bool isCoalescedLeaf =
                        (tileIsCoalesced
                            ? (coalescedPathLength == maxChildDepth)
                            : false);
                    const bool isLeaf = (in_tile.isLeaf() || in_tile.depth() == kMaxDepth) || isCoalescedLeaf;
                    // If the tile is coalesced and a leaf then we set its tileshift to 1 because it will point to the
                    // default leaf-tile with Traits::kTileSize invalid children.
                    const int tileShift = ((tileIsCoalesced && !isLeaf) ? coalescedPathLength : 0) + 1;
                    // If it's a removed leaf it means that we need to move its children pointers into the coalesced parent.
                    const bool tileIsRemovedLeaf = (tileIsRemoved && ((in_tile.depth()==kMaxDepth) || tilesToCoalesce[depth][in_tile.index()] == 0));
#else // NANOVTT_SUPPORT_COALESCED_TILES
                    const bool isLeaf = (in_tile.isLeaf() || in_tile.depth() == kMaxDepth);
                    const TileIndex accumParent =
                        (depth > rootDepth
                            ? containerVTT
                            ->m_tileCountAccum[depth - 1]
                            : 0);
#endif // !NANOVTT_SUPPORT_COALESCED_TILES
                    if (verbose)
                        if (tileIsCoalesced) ++numTilesCoalesced;
                    const TileIndex accum =
                        containerVTT->m_tileCountAccum[depth];
                    const TileIndex accumChildren =
                        containerVTT->m_tileCountAccum[depth + 1];
                    const TileIndex indexVTT =
                        accum +
                        indexToVTT[in_tile.depth()][in_tile.index()];

#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                    const BIFTile* parentTile = nullptr;
                    if (!tileIsRemoved) {
                        // Set parent.
                        // Here we need to continue until we find a parent that
                        // is not removed.
                        if (depth == rootDepth) {
                            containerVTT->setParent(indexVTT,
                                                    kInvalidTileIndex);
                        } else {
                            parentTile =
                                &acc.tile(in_tile.parentTreeIndex());
                            if (coalesceTileData) {
                                bool parentIsRemoved =
                                    tilesToRemove[parentTile->depth()]
                                                 [parentTile->index()];

                                while (parentIsRemoved) {
                                    parentTile = &acc.tile(
                                        parentTile->parentTreeIndex());
                                    parentIsRemoved =
                                        tilesToRemove[parentTile->depth()]
                                                     [parentTile->index()];
                                }
                            }
                            const TileIndex accumParent =
                                containerVTT
                                ->m_tileCountAccum[parentTile->depth()];
                            containerVTT->setParent(
                                indexVTT, accumParent +
                                              indexToVTT[parentTile->depth()]
                                                        [parentTile->index()]);
                        }
                    } // if (!tileIsRemoved)
#else // NANOVTT_SUPPORT_COALESCED_TILES
                    containerVTT->setParent(indexVTT, depth == rootDepth
                        ? kInvalidTileIndex
                        : accumParent + indexToVTT[depth - 1]
                            [in_tile.parentIndex()]);
#endif // !NANOVTT_SUPPORT_COALESCED_TILES

#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                    if (tileIsRemoved) {
                        // We only insert the child pointers into the parent if its a removed leaf
                        // otherwise we just return here because the tile has been removed and its
                        // child pointers will not be used in the NanoVTT container.
                        if (tileIsRemovedLeaf) {
                            // Here we need to find the parent coalesced tile.
                            // Note that if the parent is a coalesced leaf then
                            // nothing should be done.
                            const auto* parentTile2 =
                                &acc.tile(in_tile.parentTreeIndex());
                            bool parentIsRemoved =
                                tilesToRemove[parentTile2->depth()]
                                [parentTile2->index()];
                            while (parentIsRemoved) {
                                parentTile2 =
                                    &acc.tile(parentTile2->parentTreeIndex());
                                parentIsRemoved =
                                    tilesToRemove[parentTile2->depth()]
                                    [parentTile2->index()];
                            }
                            // The found parentTile2 must be a coalesced tile.
                            // We must first check if it's a coalesced leaf.
                            // Here we assert in debug mode if not coalesced.
                            assert(!(parentTile2->depth()==kMaxDepth) &&
                                tilesToCoalesce[parentTile2->depth()]
                                [parentTile2->index()] > 0 &&
                                (parentTile2->depth() == rootDepth ||
                                    tilesToCoalesce[parentTile2->depth() - 1]
                                    [parentTile2->parentIndex()] ==
                                    0));
                            const int coalescedPathLengthParent =
                                tilesToCoalesce[parentTile2->depth()]
                                [parentTile2->index()];
                            const int maxChildDepthParent =
                                tilesMaxChildDepth[parentTile2->depth()]
                                [parentTile2->index()];
                            const bool isCoalescedLeafParent =
                                (coalescedPathLengthParent == maxChildDepthParent);

                            // If it's a coalesced leaf parent it means that it has no
                            // children and it's not necessary to insert children into the parent.
                            if (!isCoalescedLeafParent) {
                                // Here we must insert child pointers into parent.
                                const TileIndex accumParent =
                                    containerVTT
                                        ->m_tileCountAccum[parentTile2->depth()];
                                const TileIndex parentIndexVTT =
                                    accumParent +
                                    indexToVTT[parentTile2->depth()]
                                              [parentTile2->index()];
                                TileIndex* tileVTTParent =
                                    containerVTT->children(parentIndexVTT);
                                const int twParent = static_cast<int>(std::pow(
                                    Traits::kTileWidth,
                                    (coalescedPathLengthParent + 1)));
                                const int twParent2 = twParent * twParent;
                                // The parent inherits the voxel width of the child
                                // since all children that will be moved into the parent
                                // are at the same depth.
                                const int vwParent = in_tile.voxelWidth();
                                const Bifrost::Math::int3 tileIJKParent{
                                    (in_tile.ti() - parentTile2->ti()) / vwParent,
                                    (in_tile.tj() - parentTile2->tj()) / vwParent,
                                    (in_tile.tk() - parentTile2->tk()) / vwParent
                                };
                                assert((in_tile.ti() - parentTile2->ti()) % vwParent == 0);
                                assert((in_tile.tj() - parentTile2->tj()) % vwParent == 0);
                                assert((in_tile.tk() - parentTile2->tk()) % vwParent == 0);
                                for (int k = 0; k < Traits::kTileWidth; ++k)
                                    for (int j = 0; j < Traits::kTileWidth; ++j)
                                        for (int i = 0; i < Traits::kTileWidth; ++i) {
                                            assert(tileIJKParent.x + i < twParent);
                                            assert(tileIJKParent.y + j < twParent);
                                            assert(tileIJKParent.z + k < twParent);
                                            const int indexParent =
                                                (tileIJKParent.x + i) +
                                                (tileIJKParent.y + j) *
                                                    twParent +
                                                (tileIJKParent.z + k) *
                                                    twParent2;
                                            const auto childIndex = in_tile.childIndex(i, j, k);
                                            if (childIndex >= 0) {
                                                tileVTTParent[indexParent] =
                                                    accumChildren +
                                                    indexToVTT[depth + 1]
                                                    [childIndex];
                                            }
                                            else {
                                                tileVTTParent[indexParent] = kInvalidTileIndex;
                                            }
                                        } // for (int i = 0; i < Traits::kTileWidth; ++i)
                            } // if (!isCoalescedLeafParent)
                        } // if (tileIsRemovedLeaf)
                        // Tile was removed, no more work to do.
                        return;
                    } // if (tileIsRemoved)
#endif // NANOVTT_SUPPORT_COALESCED_TILES

                    if (!(in_tile.depth() == Traits::kMaxDepth)) {
                        containerVTT->setTileOffset(indexVTT,
                            tOffset[indexVTT]);
#if defined(NANOVTT_SUPPORT_COALESCED_TILES)
                        containerVTT->setTileShift(indexVTT,
                            static_cast<uint8_t>(tileShift - 1));
#endif

#if defined(PEPTIDE_DEBUG) && defined(NANOVTT_SUPPORT_COALESCED_TILES)
                        // Check that tile shift and depth of parent agree
                        if (in_tile.depth() > containerVTT->rootDepth())
                        {
                            const TileIndex parent = containerVTT->parent(indexVTT);
                            const int tileShiftParent = containerVTT->tileShift(parent);
                            if (in_tile.depth() != parentTile->depth() + tileShiftParent) {
                                std::cout << "ERR: " << indexVTT << " " << in_tile.depth() << " " << parentTile->depth() << " " << tileShiftParent << " " << parent << std::endl;
                            }
                            assert(in_tile.depth() == parentTile->depth() + tileShiftParent);
                        }
#endif

#ifdef NANOVTT_SUPPORT_COALESCED_TILES
                        if (tileIsCoalesced) {
                            // NOP: Child pointers will be set by child tiles.
                            return;
                        }
#endif // NANOVTT_SUPPORT_COALESCED_TILES
                        if (!isLeaf) {
                            // This is a non-coalesced, non-removed, non-leaf tile.
                            TileIndex* tileVTT = containerVTT->children(indexVTT);
                            for (int i = 0; i < Traits::kTileSize; ++i) {
                                if (!tileIsCoalesced && in_tile.childIndex(i) >= 0) {
                                    tileVTT[i] =
                                        accumChildren +
                                        indexToVTT[depth + 1]
                                                  [in_tile.childIndex(i)];
                                }
                                else {
                                    tileVTT[i] = kInvalidTileIndex;
                                }
                            }
                        }
                    } // if (!(in_tile.depth()==kMaxDepth))
                },
                // Note: we need kParallelBreadthFirst (not just kParallel)
                // because parent tiles need to be set properly before the
                // children due to coalescing of tiles.
                BIF::TileTree3D::TraversalMode::kParallelBreadthFirst,
                BIF::TileTree3D::kInvalidTreeIndex, rootDepth,
                kMaxDepth);

                if (verbose)
                    std::cout << "numTilesCoalesced = " << numTilesCoalesced << std::endl;

            {
                containerVTT->setTileOffset(static_cast<TileIndex>(tOffset.size() - 1), tOffset.back());
#if defined(NANOVTT_SUPPORT_COALESCED_TILES)
                containerVTT->setTileShift(static_cast<TileIndex>(tOffset.size() - 1), 0);
#endif
                TileIndex* tileVTT = containerVTT->children(static_cast<TileIndex>(tOffset.size() - 1));
                for (int i = 0; i < Traits::kTileSize; ++i) {
                    tileVTT[i] = kInvalidTileIndex;
                }
            }
            if (verbose) {
                containerSizeOpt += numNonLeavesAtCoarserDepths * sizeof(TileIndex) * Traits::kTileSize;
                std::cout << "numNonLeavesAtCoarserDepths = " << numNonLeavesAtCoarserDepths << std::endl;
                std::cout << "Container size = " << static_cast<float>(containerSize) / 1024 / 1024 << "MB" << std::endl;
                std::cout << "Container hash table size = " << static_cast<float>(Container::kTableSize * sizeof(uint32_t)) / 1024 / 1024 << "MB" << std::endl;
                std::cout << "Potential container memory savings = " << static_cast<float>(containerSize - containerSizeOpt) / 1024 / 1024 << "MB"
                    << " (" << static_cast<float>(containerSize - containerSizeOpt) / containerSize * 100 << "%)" << std::endl;
                std::cout << "Potential total memory savings: before: " << bufferSize / 1024 / 1024 << "MB, after: " <<
                    (bufferSize - containerSize + containerSizeOpt) / 1024 / 1024 << "MB "
                    << "(" << static_cast<float>(containerSize - containerSizeOpt) / bufferSize * 100 << "%)" << std::endl;
            }
            endTime2 = tbb::tick_count::now();
            if (verbose || verboseTimings) {
                std::cout << "Container pass 2 = "
                    << (endTime2 - startTime2).seconds() << " seconds" << std::endl;
            }

            // CONTAINER.END (pass 2)

            // CHANNELS.BEGIN (pass 2)

            startTime2 = tbb::tick_count::now();

            // Initialize grid pointer
            uint8_t* gridVTTBase =
                reinterpret_cast<uint8_t*>(containerVTT) + containerSize;
            uint32_t gridCounter = 0;

            Amino::for_each<NanoVTTTypes>([&](auto in) {
                using PairType = decltype(in);
                using VTTType = std::tuple_element_t<0, PairType>;
                using NanoVTTType = std::tuple_element_t<1, PairType>;
                using BuildType = typename std::conditional<
                    std::is_same<NanoVTTType, float>::value && !std::is_same<VoxelT, void>::value,
                    VoxelT, NanoVTTType>::type;
                using BIFChannel =
                    BIF::TileTree3D::Channel<BIFTraits, VTTType,
                                             BIFTraits::kTileDataWidth>;
                using TDH = BIF::TileTree3D::TileData<VTTType, Traits::kTileDataWidth + 1>;

                for (const Amino::String& name : in_volume.componentRange()) {
                    if (!parser.accepts(name)) continue;
                    if (auto channel = in_volume.component<BIFChannel>(name)) {
                        if (verbose) {
                            std::cout << "Channel " << name.c_str() << " stats:" << std::endl;
                            std::cout << " - minDepth = " << channel->minDepth() << std::endl;
                            std::cout << " - maxDepth = " << channel->depth() << std::endl;
                        }

                        const bool isSDF = ::BIF::TileTree3D::LevelSet::isSignedDistance(*channel);
                        const auto defaultValue = channel->defaultValue();

                        // Here we need to copy channel and set coarser values if required.
                        // Note: could potentially be optimized by only copying and allocating
                        //       coarser levels.
                        tbb::tick_count startTime3, endTime3;
                        startTime3 = tbb::tick_count::now();
                        if (setCoarserLevels) {

                            const float epsilon = in_activeFogVoxelsThreshold;
                            const bool areCoarserOK = BIF::TileTree3D::checkIfNonLeafValueIsSetToChildValue(*channel,
                                *container,
                                voxelTopology->offsetType(),
                                epsilon,
                                rootDepth,
                                BIFTraits::kMaxDepth);

                            if (verbose) {
                                std::cout << "areCoarserOK = " << areCoarserOK << std::endl;
                            }

                            if (!areCoarserOK) {
                                Amino::Ptr<BIFChannel> tmpChannel;
                                if (BIF::TileTree3D::ChannelPool::enabled()) {
                                    tmpChannel =
                                        BIF::TileTree3D::ChannelPool::allocate<
                                        BIFTraits, VTTType,
                                        BIFTraits::kTileDataWidth, /*outH*/ 0>(
                                            defaultValue,
                                            /*useStaticAllocation*/ false,
                                            /*callElementConstructor*/ true,
                                            rootDepth, channel->depth(),
                                            voxelTopology->tileCount(),
                                            voxelTopology.get());
                                    temp_const_cast(*tmpChannel).setName(name);
                                }
                                if (tmpChannel == nullptr) {
                                    tmpChannel =
                                        Amino::newClassPtr<BIFChannel>(
                                            name,
                                            &defaultValue,
                                            BIF::TileTree3D::
                                            DataElementInitializerType::
                                            kDefaultValue,
                                            /*useStaticAllocation*/ false,
                                            /*callElementConstructor*/ true,
                                            rootDepth, channel->depth());
                                }
                                temp_const_cast(*tmpChannel).update(*voxelTopology);
                                container->traverse(
                                    [&](const auto& in_tile) {
                                        auto in2 =
                                            channel->tileData(in_tile.treeIndex());
                                        auto out2 =
                                            temp_const_cast(*tmpChannel)
                                            .tileData(in_tile.treeIndex());
                                        out2.setCore(in2);
                                    },
                                    BIF::TileTree3D::TraversalMode::kParallel,
                                        BIF::TileTree3D::kInvalidTreeIndex,
                                        rootDepth, channel->depth());
                                BIF::TileTree3D::setNonLeafValueToChildValue(temp_const_cast(*tmpChannel),
                                    *container,
                                    voxelTopology->offsetType(),
                                    rootDepth,
                                    channel->depth());
                                channel = tmpChannel;
                            } // if (!areCoarserOK)
                        } // setCoarserLevels
                        endTime3 = tbb::tick_count::now();
                        if (verbose || verboseTimings) {
                            std::cout << name.c_str() << ": Setting coarser levels = "
                                << (endTime3 - startTime3).seconds() << " seconds" << std::endl;
                        }

                        // Grid and Tree
                        Grid<Tree<BuildType>>* gridVTT =
                            reinterpret_cast<Grid<Tree<BuildType>>*>(gridVTTBase);
                        Tree<BuildType> *treeVTT = &gridVTT->tree();
                        constexpr uint64_t valueTypeSize = sizeof(typename Tree<BuildType>::ValueType);

                        // Tree
                        treeVTT->m_containerOffset =
                            reinterpret_cast<uint8_t*>(treeVTT) -
                            reinterpret_cast<uint8_t*>(containerVTT);
                        treeVTT->m_treeSize = gridAndTreeSize[gridCounter] - sizeof(Grid<Tree<BuildType>>);
                        if (!isSDF || !kClampSDFVoxelValues) {
                            treeVTT->m_defaultValue = *reinterpret_cast<const NanoVTTType*>(&defaultValue);
                        }
                        else {
                            const VTTType defaultSDFValue = clampSDFDefaultValue(defaultValue, voxelSize);
                            treeVTT->m_defaultValue = *reinterpret_cast<const NanoVTTType*>(&defaultSDFValue);
                        }

                        // Grid
                        nanovdb::GridData* gridDataVTT = reinterpret_cast<nanovdb::GridData*>(gridVTT);
                        gridDataVTT->setLongGridNameOn(false);
                        gridDataVTT->mMagic = NANOVTT_MAGIC_NUMBER;
                        gridDataVTT->mVersion =
                            nanovdb::Version(NANOVTT_MAJOR_VERSION_NUMBER,
                                NANOVTT_MINOR_VERSION_NUMBER,
                                NANOVTT_PATCH_VERSION_NUMBER);
                        if (isSDF) {
                            gridDataVTT->mGridClass = nanovdb::GridClass::LevelSet;
                        }
                        else if (name == Bifrost::Geometry::sVoxelFogDensity) {
                            gridDataVTT->mGridClass = nanovdb::GridClass::FogVolume;
                        }
                        else {
                            gridDataVTT->mGridClass = nanovdb::GridClass::Unknown;
                        }
                        gridDataVTT->mGridIndex = gridCounter;
                        gridDataVTT->mGridCount = gridCount;
                        gridDataVTT->mGridSize = gridAndTreeSize[gridCounter];
                        strncpy(gridDataVTT->mGridName, name.c_str(),
                            nanovdb::GridData::MaxNameSize - 1);
                        gridDataVTT->mGridName
                            [nanovdb::GridData::MaxNameSize - 1] = 0;
                        if (verbose)
                            std::cout << "name = '" << name.c_str() << "'" << std::endl;
                        gridDataVTT->mGridType = nanovdb::mapToGridType<BuildType>();
                        gridDataVTT->mVoxelSize[0] = static_cast<double>(voxelSize);
                        gridDataVTT->mVoxelSize[1] = static_cast<double>(voxelSize);
                        gridDataVTT->mVoxelSize[2] = static_cast<double>(voxelSize);

                        gridDataVTT->mBlindMetadataOffset = 0;
                        gridDataVTT->mBlindMetadataCount = 0;
                        gridDataVTT->setLongGridNameOn(false);

                        const double Sx = static_cast<double>(voxelSize), Sy = static_cast<double>(voxelSize), Sz = static_cast<double>(voxelSize);
                        constexpr double Tx = 0, Ty = 0, Tz = 0;
                        const double SxInv = 1 / Sx, SyInv = 1 / Sy, SzInv = 1 / Sz;
                        const double mat[4][4] = {
                            {Sx, 0.0, 0.0, 0.0}, // row 0
                            {0.0, Sy, 0.0, 0.0}, // row 1
                            {0.0, 0.0, Sz, 0.0}, // row 2
                            {Tx, Ty, Tz, 1.0},          // row 3
                        };
                        const double invMat[4][4] = {
                            {SxInv , 0.0, 0.0, 0.0}, // row 0
                            {0.0, SyInv, 0.0, 0.0}, // row 1
                            {0.0, 0.0, SzInv, 0.0}, // row 2
                            {-Tx, -Ty, -Tz, 1.0},           // row 3
                        };
                        gridDataVTT->mMap.set(mat, invMat, 1.0);

                        if (verbose) {
                            std::cout << "containerOffset = "
                                << treeVTT->m_containerOffset
                                << std::endl;
                            std::cout << "Size of tile data = "
                                << Tree<BuildType>::kTileDataBytes << std::endl;
                        }

                        using TD = BIF::TileTree3D::ConstTileData<VTTType, Traits::kTileDataWidth>;

                        // The lambda-approach does not compile with VS2019 so
                        // we need to construct a visitor.
                        Visitor<VoxelT, BuildType, VTTType, TD, TDH, valueTypeSize> visitor(treeVTT, lut, acc, containerVTT,
                                        tdOffset, tdOffsetConstant, indexToVTT,
                                        tdcOffset, tilesToCoalesce, rootDepth,
                                        kMaxDepth, coalesceTileData, tilesToRemove,
                                        *channel, codecsVec, fpNIndex,
                                        tdOffsetFpNVec, isSDF, container->voxelSize(),
                                        in_sortIntoMortonOrder, permutation,
                                        in_dilateActiveFogVoxels, in_activeFogVoxelsThreshold,
                                        texture3DDim, storeTileDataAs3DTextures);

                        {
                            startTime3 = tbb::tick_count::now();
                            container->reduce(
                                &visitor,
                                BIF::TileTree3D::TraversalMode::kParallelBreadthFirst,
                                BIF::TileTree3D::kInvalidTreeIndex,
                                rootDepth, kMaxDepth);
                            endTime3 = tbb::tick_count::now();
                            if (verbose || verboseTimings) {
                                std::cout << name.c_str() << ": Adding data to tiles and computing bbox = "
                                    << (endTime3 - startTime3).seconds() << " seconds" << std::endl;
                            }

#ifdef PEPTIDE_DEBUG
                            if (name == Bifrost::Geometry::sVoxelFogDensity &&
                                in_extendedBudgetFraction >= 1.f &&
                                in_dilateActiveFogVoxels &&
                                std::is_same<VoxelT, void>::value && !coalesceTileData &&
                                !in_storeTileDataAs3DTextures) {
                                testTileData(*container, *treeVTT,
                                             tileCountAccum, indexToVTT,
                                             in_activeFogVoxelsThreshold,
                                             rootDepth, kMaxDepth);
                            }
#endif
                        }

                        // Set entry 'numNanoVTTTiles' to facilitate
                        // computation of sizes of tiles.
                        if (std::is_same<VoxelT, void>::value || !std::is_same<VTTType, float>::value) {
                            treeVTT->setTileDataOffset(
                                static_cast<TileIndex>(numNanoVTTTiles),
                                numHaloTiles * valueTypeSize *
                                        Traits::kExtendedTileDataSize +
                                    numNonHaloTiles * valueTypeSize *
                                        Traits::kTileDataSize +
                                    tdOffsetConstant);
                        } else if (!std::is_same<VoxelT, nanovdb::FpN>::value) {
                            treeVTT->setTileDataOffset(
                                static_cast<TileIndex>(numNanoVTTTiles),
                                numHaloTiles *
                                        Tree<BuildType>::kExtendedTileDataBytes +
                                    numNonHaloTiles *
                                        Tree<BuildType>::kTileDataBytes +
                                    tdOffsetConstant);
                        }
                        if (std::is_same<VoxelT, nanovdb::FpN>::value && std::is_same<VTTType, float>::value) {
                            ++fpNIndex;
                            if (verbose) {
                                std::cout << "Incrementing fpNIndex = " << fpNIndex << std::endl;
                            }
                        }

                        gridDataVTT->mWorldBBox[0][0] = static_cast<double>(visitor.m_imin * voxelSize);
                        gridDataVTT->mWorldBBox[0][1] = static_cast<double>(visitor.m_jmin * voxelSize);
                        gridDataVTT->mWorldBBox[0][2] = static_cast<double>(visitor.m_kmin * voxelSize);
                        gridDataVTT->mWorldBBox[1][0] = static_cast<double>(visitor.m_imax * voxelSize);
                        gridDataVTT->mWorldBBox[1][1] = static_cast<double>(visitor.m_jmax * voxelSize);
                        gridDataVTT->mWorldBBox[1][2] = static_cast<double>(visitor.m_kmax * voxelSize);

                        if (out_bbox && visitor.m_enableActiveState) {
                            *out_bbox = Bifrost::Math::bboxf(
                                {visitor.m_imintf * voxelSize,
                                 visitor.m_jmintf * voxelSize,
                                 visitor.m_kmintf * voxelSize},
                                {visitor.m_imaxtf * voxelSize,
                                 visitor.m_jmaxtf * voxelSize,
                                 visitor.m_kmaxtf * voxelSize});
                        }

                        treeVTT->m_indexBBox[0][0] = visitor.m_imin;
                        treeVTT->m_indexBBox[0][1] = visitor.m_jmin;
                        treeVTT->m_indexBBox[0][2] = visitor.m_kmin;
                        treeVTT->m_indexBBox[1][0] = visitor.m_imax;
                        treeVTT->m_indexBBox[1][1] = visitor.m_jmax;
                        treeVTT->m_indexBBox[1][2] = visitor.m_kmax;
                        if (verbose)
                        {
                            std::cout << "WorldBBox = ("
                                      << gridDataVTT->mWorldBBox[0][0] << ", "
                                      << gridDataVTT->mWorldBBox[0][1] << ", "
                                      << gridDataVTT->mWorldBBox[0][2] << ") to ("
                                      << gridDataVTT->mWorldBBox[1][0] << ", "
                                      << gridDataVTT->mWorldBBox[1][1] << ", "
                                      << gridDataVTT->mWorldBBox[1][2] << ")"
                                      << std::endl;
                        } // if (verbose)


                        gridDataVTT->mChecksum = checksum(*gridVTT, in_cMode);

                        gridVTTBase += gridAndTreeSize[gridCounter];
                        if (verbose) {
                            std::cout << "Adding gridAndTreeSize " << gridAndTreeSize[gridCounter] << std::endl;
                            std::cout << "GridVTTBase = " << reinterpret_cast<uint64_t>(gridVTTBase) << std::endl;
                            std::cout << "GridCount = " << gridVTT->gridCount() << std::endl;
                            std::cout << "GridIndex = " << gridVTT->gridIndex() << std::endl;
                            std::cout << "GridSize = " << gridVTT->gridSize() << std::endl;
                        }
                        ++gridCounter;

                        if (verbose || verboseTimings) {
                            std::cout << name.c_str() << ": Setting to nullptr (potential dealloc) = ";
                        }
                        startTime3 = tbb::tick_count::now();
                        channel = nullptr;
                        endTime3 = tbb::tick_count::now();
                        if (verbose || verboseTimings) {
                            std::cout << (endTime3 - startTime3).seconds() << " seconds" << std::endl;
                        }

                    }
                }
            });   // Amino::for_each<NanoVTTTypes>([&](auto in)

            // CHANNELS.END (pass 2)

        }         // if (auto voxelTopology =
          // in_volume.component<BIFVoxelTopology>(Bifrost::Geometry::sVoxelComp))
        endTime2 = tbb::tick_count::now();
        if (verbose || verboseTimings) {
            std::cout << "Channels pass 2 = "
                << (endTime2 - startTime2).seconds() << " seconds" << std::endl;
        }

    endTime = tbb::tick_count::now();
    if (verbose || verboseTimings)
    {
        std::cout << "converted from VTT to NanoVTT in "
                  << (endTime - startTime).seconds() << " seconds" << std::endl;
    }

#if 0
    std::cout << "End profiling" << std::endl;
    std::this_thread::sleep_for(std::chrono::milliseconds(20000));
    std::cout << "Profile ended" << std::endl;
#endif

    return handle;
} // ... vttToNanoVTT(...)

} // namespace nanovtt
